# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains helper functions for Bash.'''

from .helpers import GeneralHelpers, ShellFunction

_VALUES = ShellFunction('values', r'''
local word append=0

[[ "$1" == "-a" ]] && { shift; append=1; }
[[ "$1" == "--" ]] && { shift; }

(( append )) || COMPREPLY=()

for word; do
  if [[ "$word" == "$cur"* ]]; then
    COMPREPLY+=("$(printf '%q' "$word")")
  fi
done
''')

_DEQUOTE = ShellFunction('dequote', r'''
local in="$1" len=${#1} i=0 result='' ___break_pos=-1 ___in_quotes=0

for ((; i < len; ++i)); do
  case "${in:i:1}" in
    "'")
      ___in_quotes=1
      for ((++i; i < len; ++i)); do
        [[ "${in:i:1}" == "'" ]] && { ___in_quotes=0; break; }
        result+="${in:i:1}"
      done;;
    '"')
      ___in_quotes=1
      for ((++i; i < len; ++i)); do
        [[ "${in:i:1}" == '"' ]] && { ___in_quotes=0; break; }

        if [[ "${in:i:1}" == '\' ]]; then
          result+="${in:$((++i)):1}"
        else
          result+="${in:i:1}"
        fi
      done;;
    '\')
      result+="${in:$((++i)):1}";;
    [$COMP_WORDBREAKS])
      result+="${in:i:1}"
      ___break_pos=${#result};;
    *)
      result+="${in:i:1}";;
  esac
done

local -n ___RESULT=$2
local -n ___BREAK_POS=$3
local -n ___IN_QUOTES=$4
___RESULT="$result"
___BREAK_POS=$___break_pos
___IN_QUOTES=$___in_quotes
''')

_DEQUOTE_WORDS = ShellFunction('dequote_words', r'''
local word dequoted break_pos in_quotes

words_dequoted=()

for word in "${words[@]}"; do
  dequote "$word" dequoted break_pos in_quotes
  words_dequoted+=("$dequoted")
done
''', ['dequote'])

_PARSE_LINE = ShellFunction('parse_line', r'''
local -n out="$1"
local in="$2" len=${#2} i=0 word='' active=0

if (( len == 0 )); then
  out+=('')
  return
fi

for ((i=0; i < len; ++i)); do
  case "${in:i:1}" in
    ' ')
      if (( active )); then
        out+=("$word")
        word=''
        active=0;
      fi;;
    '"')
      for ((; i < len; ++i)); do
        if [[ "${in:i:1}" == '\' ]]; then
          word+="${in:$((++i)):1}"
        else
          word+="${in:i:1}"
          [[ "${in:i:1}" == '"' ]] && break;
        fi
      done
      active=1;;
    "'")
      for ((; i < len; ++i)); do
        word+="${in:i:1}"
        [[ "${in:i:1}" == "'" ]] && break;
      done
      active=1;;
    '\')
      word+='\'
      word+="${in:$((++i)):1}"
      active=1;;
    *)
      word+="${in:i:1}"
      active=1;;
  esac
done

if (( active )); then
  out+=("$word")
elif [[ "${in:$((len - 1)):1}" == ' ' ]]; then
  out+=('')
fi
''')

_GET_PREFIX_SUFFIX_LEN = ShellFunction('get_prefix_suffix_len', r'''
local s1="$1"
local s2="$2"

local len1=${#s1}
local len2=${#s2}

local maxk=$len1
if (( len2 < maxk )); then
  maxk=$len2
fi

local k=$maxk
while (( k > 0 )); do
  local start=$((len1 - k))
  local suf="${s1:start:k}"
  local pre="${s2:0:k}"

  if [[ "$suf" == "$pre" ]]; then
    if (( len1 == k )); then
      echo 0
    else
      echo $((len1-k))
    fi

    return
  fi

  ((k--))
done

echo ${#s1}
''')

_COMMANDLINE_STRING = ShellFunction('commandline_string', r'''
local l line break_pos in_quotes
dequote "$cur" line break_pos in_quotes

local COMP_LINE_OLD="$COMP_LINE"
local COMP_POINT_OLD=$COMP_POINT
local COMP_WORDS_OLD=("${COMP_WORDS[@]}")
local COMP_CWORD_OLD=$COMP_CWORD

COMP_LINE="$line"
COMP_POINT=${#line}
COMP_WORDS=()
parse_line COMP_WORDS "$line"
COMP_CWORD=$(( ${#COMP_WORDS[@]} - 1 ))
COMPREPLY=()

#ifdef bash_completions_v_2_12
_comp_command_offset 0
#else
_command_offset 0
#endif
compopt -o nospace
compopt -o noquote

if (( break_pos >= 0 )); then
  line="${line:break_pos}"
fi

if (( ${#COMPREPLY[@]} )); then
  local len=$(get_prefix_suffix_len "$line" "${COMPREPLY[0]}")
  line="${line:0:len}"

  for i in "${!COMPREPLY[@]}"; do
    local REPLY="${COMPREPLY[i]}"
    if (( in_quotes )); then
      COMPREPLY[i]="$line$REPLY"
    else
      COMPREPLY[i]="$(printf '%q' "$line$REPLY")"
    fi
  done
fi

COMP_LINE="$COMP_LINE_OLD"
COMP_POINT=$COMP_POINT_OLD
COMP_WORDS=("${COMP_WORDS_OLD[@]}")
COMP_CWORD=$COMP_CWORD_OLD
''', ['dequote', 'parse_line', 'get_prefix_suffix_len'])

_EXEC = ShellFunction('exec', r'''
local item desc special="$COMP_WORDBREAKS\"'><=;|&({:\\\$\`"

while IFS=$'\t' read -r item desc; do
  if [[ "$item" == "$cur"* ]]; then
    [[ "$item" == *[$special]* ]] && item="$(printf '%q' "$item")"
    COMPREPLY+=("$item")
  fi
done < <(eval "$1")
''')

_EXEC_FAST = ShellFunction('exec_fast', r'''
local item desc

while IFS=$'\t' read -r item desc; do
  if [[ "$item" == "$cur"* ]]; then
    COMPREPLY+=("$item")
  fi
done < <(eval "$1")
''')

_ARRAY_CONTAINS = ShellFunction('array_contains', r'''
local w='' search="$1"; shift;
for w; do [[ "$search" == "$w" ]] && return 0; done
return 1
''')

_VALUE_LIST = ShellFunction('value_list', r'''
#ifdef duplicates
local duplicates=0
[[ "$1" == '-d' ]] && { duplicates=1; shift; }
#endif
local separator="$1"; shift

compopt -o nospace

local cur_unquoted break_pos in_quotes
dequote "$cur" cur_unquoted break_pos in_quotes

if [[ -z "$cur_unquoted" ]]; then
  COMPREPLY=("$@")
  return
fi

local value having_value having_values=() remaining_values=()

IFS="$separator" read -r -a having_values <<< "$cur_unquoted"

#ifdef duplicates
if (( duplicates )); then
  remaining_values=("$@")
else
  for value; do
    if ! array_contains "$value" "${having_values[@]}"; then
      remaining_values+=("$value")
    fi
  done
fi
#else
for value; do
  if ! array_contains "$value" "${having_values[@]}"; then
    remaining_values+=("$value")
  fi
done
#endif

COMPREPLY=()

local cur_stripped="$cur_unquoted"
if (( break_pos > -1 )); then
  cur_stripped="${cur_stripped:break_pos}"
fi

if [[ "${cur_unquoted: -1}" == "$separator" ]]; then
  for value in "${remaining_values[@]}"; do
    COMPREPLY+=("$cur_stripped$value")
  done
elif (( ${#remaining_values[@]} )); then
  if array_contains "${having_values[-1]}" "$@"; then
    COMPREPLY+=("$cur_stripped$separator")
  elif (( ${#having_values[@]} )); then
    local cur_last_value=${having_values[-1]}
    cur_stripped="${cur_stripped%"$cur_last_value"}"

    for value in "${remaining_values[@]}"; do
      if [[ "$value" == "$cur_last_value"* ]]; then
        COMPREPLY+=("$cur_stripped$value")
      fi
    done
  fi
fi
''', ['dequote', 'array_contains'])

_KEY_VALUE_LIST = ShellFunction('key_value_list', r'''
local sep1="$1"; shift
local sep2="$1"; shift
local -A keys=()
local i

for ((i=1; i <= $#; i += 2)); do
  keys["${@:i:1}"]="${@:i + 1:1}"
done

local strip_chars=''
[[ "$COMP_WORDBREAKS" == *"$sep1"* ]] && strip_chars+="$sep1"
[[ "$COMP_WORDBREAKS" == *"$sep2"* ]] && strip_chars+="$sep2"
[[ "${cur:0:1}" == '"' ]] && strip_chars=''
[[ "${cur:0:1}" == "'" ]] && strip_chars=''
local cur="$cur" break_pos in_quotes
dequote "$cur" cur break_pos in_quotes

if [[ -z "$cur" ]]; then
  COMPREPLY=("${!keys[@]}")
  return
fi

local pair key value found_key cur_stripped="$cur"
local -a having_pairs having_keys remaining_keys

IFS="$sep1" read -r -a having_pairs <<< "$cur"

for pair in "${having_pairs[@]}"; do
  having_keys+=("${pair%%"$sep2"*}")
done

for key in "${!keys[@]}"; do
  found_key=0

  for having_key in "${having_keys[@]}"; do
    if [[ "$key" == "$having_key" ]]; then
      found_key=1
      break
    fi
  done

  if (( ! found_key )); then
    remaining_keys+=("$key")
  fi
done

COMPREPLY=()

if [[ "${cur: -1}" == "$sep1" ]]; then
  [[ -n "$strip_chars" ]] && cur_stripped="${cur_stripped##*[$strip_chars]}"

  for key in "${remaining_keys[@]}"; do
    COMPREPLY+=("$cur_stripped$key")
  done
else
  pair="${cur##*"$sep1"}"
  if [[ "$pair" == *"$sep2"* ]]; then
    key="${pair%%"$sep2"*}"
    value="${pair##*"$sep2"}"
    cur="$value"
    ${keys[$key]}

    cur_stripped="${cur_stripped:0:$(( ${#cur_stripped} - ${#value} ))}"
    if [[ -n "$strip_chars" ]]; then
        cur_stripped="${cur_stripped##*[$strip_chars]}"
    fi

    for i in "${!COMPREPLY[@]}"; do
      COMPREPLY[i]="$cur_stripped${COMPREPLY[i]}"
    done
  else
    [[ -n "$strip_chars" ]] && cur_stripped="${cur_stripped##*[$strip_chars]}"
    cur_stripped="${cur_stripped%"$pair"}"

    for key in "${remaining_keys[@]}"; do
      if [[ "$key" == "$pair"* ]]; then
        COMPREPLY+=("$cur_stripped$key")
      fi
    done
  fi
fi
''', ['dequote'])

_PREFIX_COMPREPLY = ShellFunction('prefix_compreply', r'''
[[ "$cur" == *[$COMP_WORDBREAKS]* ]] && return

local i prefix="$1"
for ((i=0; i < ${#COMPREPLY[@]}; ++i)); do
  COMPREPLY[i]="$prefix${COMPREPLY[i]}"
done
''')

_HISTORY = ShellFunction('history', r'''
[[ -f "$HISTFILE" ]] || return

local match

while read -r match; do
  if [[ "$match" == "$cur"* ]]; then
    COMPREPLY+=("$(printf '%q' "$match")")
  fi
done < <(command grep -E -o -- "$1" "$HISTFILE")
''')

_GET_LAST_BREAK_POSITION = ShellFunction('get_last_break_position', r'''
local in="$1" i=0 len=${#1} pos=-1

for ((i=0; i < len; ++i)); do
  case "${in:i:1}" in
    '"')
      for ((++i; i < len; ++i)); do
        if [[ "${in:i:1}" == '\' ]]; then
          ((++i))
        elif [[ "${in:i:1}" == '"' ]]; then
          break;
        fi
      done;;
    "'")
      for ((++i; i < len; ++i)); do
        [[ "${in:i:1}" == "'" ]] && break;
      done;;
    '\')
      ((++i));;
    [$COMP_WORDBREAKS])
      pos=$i;;
  esac
done

echo $pos
''')

_PREFIX = ShellFunction('prefix', r'''
local prefix="$1" func="$2"
local stripped="$(strip_prefix_keep_quoting "$prefix" "$cur")"

if [[ "$stripped" == "$cur" ]]; then
  COMPREPLY=($(compgen -W "$1" -- "$cur"))
  compopt -o nospace
  return
fi

local break_position=$(get_last_break_position "$cur")

local cur_old="$cur" cur="$stripped"
$func

if (( break_position < 0 )); then
  for i in "${!COMPREPLY[@]}"; do
    COMPREPLY[i]="$prefix${COMPREPLY[i]}"
  done
fi
''', ['get_last_break_position', 'strip_prefix_keep_quoting'])

_STRIP_PREFIX_KEEP_QUOTING = ShellFunction('strip_prefix_keep_quoting', r'''
local p="$1" pi=0 plen=${#1}
local s="$2" si=0 slen=${#2} state=w

for (( pi=0; pi < plen; ++pi )); do
  pc=${p:pi:1}
  sc=''

  while (( si < slen )); do
    case "$state" in
      w)
        case "${s:si:1}" in
          '"') state=dq; ((++si));;
          "'") state=sq; ((++si));;
          '\')           ((++si)); sc="${s:si:1}"; ((++si));   break;;
          *)             sc="${s:si:1}"; ((++si));             break;;
        esac;;
      sq)
        case "${s:si:1}" in
          "'") state=w;  ((++si));;
          *)             sc="${s:si:1}"; ((++si));             break;;
        esac;;
      dq)
        case "${s:si:1}" in
          '"') state=w;  ((++si));;
          '\')           ((++si)); sc="${s:si:1}"; ((++si));   break;;
          *)             sc="${s:si:1}"; ((++si));             break;;
        esac;;
    esac
  done

  [[ "$pc" != "$sc" ]] && { echo "$s"; return; }
done

case "$state" in
  w)  printf '%s\n' "${s:si}";;
  sq) printf '%s\n' "'${s:si}";;
  dq) printf '%s\n' "\"${s:si}";;
esac
''')

_MIME_FILE = ShellFunction('mime_file', r'''
local line file mime i_opt cur_dequoted

if command file -i /dev/null &>/dev/null; then
  i_opt="-i"
elif command file -I /dev/null &>/dev/null; then
  i_opt="-I"
else
#ifdef bash_completions_v_2_12
  _comp_compgen_filedir
#else
  _filedir
#endif
  return
fi

local cur_dequoted break_pos in_quotes
dequote "$cur" cur_dequoted break_pos in_quotes

while read -r line; do
  mime="${line##*:}"

  if [[ "$mime" == *inode/directory* ]] || command grep -q -E -- "$1" <<< "$mime"; then
    file="${line%:*}"

    if [[ "$file" == *\\* ]]; then
      file="$(command perl -pe 's/\\([0-7]{3})/chr(oct($1))/ge' <<< "$file")"
    fi

    if [[ "$mime" == *inode/directory* ]]; then
      file="$file/"
    fi

    if [[ "$file" == "$cur_dequoted"* ]]; then
      COMPREPLY+=("$(printf '%q' "$file")")
    fi
  fi
done < <(command file -L $i_opt -- "$cur_dequoted"* 2>/dev/null)
''', ['dequote'])

_FILE_FILTER = ShellFunction('file_filter', '''
local pattern REPLY COMPREPLY_OLD=("${COMPREPLY[@]}")

COMPREPLY=()

for REPLY in "${COMPREPLY_OLD[@]}"; do
  for pattern; do
    [[ "${REPLY##*/}" == $pattern ]] && continue 2
  done

  COMPREPLY+=("$REPLY")
done
''')

_NUMBER = ShellFunction('number', '''
local number="$(command grep -E -o '^[0-9,.]+' <<< "$cur")" suffix_and_desc

[[ -z "$number" ]] && return

COMPREPLY=("$number")

for suffix_and_desc; do
  COMPREPLY+=("${number}${suffix_and_desc%%:*}")
done
''')

# =============================================================================
# Bonus
# =============================================================================

_MOUNTPOINT = ShellFunction('mountpoint', r'''
local REPLY

COMPREPLY=()

while read -r REPLY; do
  COMPREPLY+=("$(printf '%q' "$REPLY")")
done < <(command findmnt -lno TARGET)
''')

_LOCALES = ShellFunction('''locales''', r'''
COMPREPLY=($(compgen -W "$(command locale -a)" -- "$cur"))
''')

_CHARSETS = ShellFunction('''charsets''', r'''
COMPREPLY=($(compgen -W "$(command locale -m)" -- "$cur"))
''')

_TIMEZONE_LIST = ShellFunction('timezone_list', r'''
if ! command timedatectl list-timezones 2>/dev/null; then
  command find /usr/share/zoneinfo -type f |\
  command sed 's|/usr/share/zoneinfo/||g'  |\
  command grep -E -v '^(posix|right)'
fi''')

_ALSA_LIST_CARDS = ShellFunction('alsa_list_cards', r'''
local card id cards=()
while builtin read card; do
  card="${card#card }"
  id="${card%%: *}"
  cards+=("$id")
done < <(command aplay -l | command grep -Eo '^card [0-9]+: [^,]+')

COMPREPLY=($(compgen -W "${cards[*]}" -- "$cur"))
''')

_ALSA_LIST_DEVICES = ShellFunction('alsa_list_devices', r'''
local card id devices=()
while builtin read card; do
  card="${card#card }"
  id="${card%%: *}"
  devices+=("hw:$id")
done < <(command aplay -l | command grep -Eo '^card [0-9]+: [^,]+')

COMPREPLY=($(compgen -W "${devices[*]}" -- "$cur"))
''')


class BashHelpers(GeneralHelpers):
    '''Class holding helper functions for Bash.'''

    def __init__(self, config, function_prefix):
        super().__init__(config, function_prefix, ShellFunction)
        self.add_function(_VALUES)
        self.add_function(_DEQUOTE)
        self.add_function(_DEQUOTE_WORDS)
        self.add_function(_PREFIX)
        self.add_function(_STRIP_PREFIX_KEEP_QUOTING)
        self.add_function(_ARRAY_CONTAINS)
        self.add_function(_GET_LAST_BREAK_POSITION)
        self.add_function(_PARSE_LINE)
        self.add_function(_GET_PREFIX_SUFFIX_LEN)
        self.add_function(_COMMANDLINE_STRING)
        self.add_function(_EXEC)
        self.add_function(_EXEC_FAST)
        self.add_function(_VALUE_LIST)
        self.add_function(_KEY_VALUE_LIST)
        self.add_function(_FILE_FILTER)
        self.add_function(_PREFIX_COMPREPLY)
        self.add_function(_HISTORY)
        self.add_function(_NUMBER)
        self.add_function(_MIME_FILE)
        self.add_function(_MOUNTPOINT)
        self.add_function(_LOCALES)
        self.add_function(_CHARSETS)
        self.add_function(_TIMEZONE_LIST)
        self.add_function(_ALSA_LIST_CARDS)
        self.add_function(_ALSA_LIST_DEVICES)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains code for validating completing definitions.'''

# Unlike the `scheme_validator`, this validator operates directly on Python
# objects and therefore does not have access to line and column information.
# To save code and maintain consistency, we reuse the validation logic
# from `scheme_validator` by wrapping the objects in `ValueWithoutTrace`.
# This provides the same structure as used in `scheme_validator`, except that
# we use `ValueWithoutTrace` instead of `ValueWithTrace`.

from types import NoneType

from . import scheme_validator
from .value_with_trace import ValueWithOutTrace


def _mk_value_without_trace(value):
    if isinstance(value, ValueWithOutTrace):
        return value

    if isinstance(value, (NoneType, str, int, float, bool)):
        return ValueWithOutTrace(value)

    mk = _mk_value_without_trace

    if isinstance(value, (list, tuple)):
        return ValueWithOutTrace(list(map(mk, value)))

    if isinstance(value, dict):
        return ValueWithOutTrace({mk(k): mk(v) for k, v in value.items()})

    raise AssertionError(f"Not reached: {value!r}")


def _option_to_definition(option):
    return {
        'option_strings': option.option_strings,
        'metavar':        option.metavar,
        'help':           option.help,
        'complete':       option.complete,
        'nosort':         option.nosort,
        'groups':         option.groups,
        'optional_arg':   option.optional_arg,
        'repeatable':     option.repeatable,
        'final':          option.final,
        'hidden':         option.hidden,
        'when':           option.when,
        'capture':        option.capture,
    }


def _positional_to_definition(positional):
    return {
        'number':     positional.number,
        'metavar':    positional.metavar,
        'help':       positional.help,
        'repeatable': positional.repeatable,
        'complete':   positional.complete,
        'nosort':     positional.nosort,
        'when':       positional.when,
        'capture':    positional.capture,
    }


def _commandline_to_definition(cmdline):
    options = list(map(_option_to_definition, cmdline.options))
    positionals = list(map(_positional_to_definition, cmdline.positionals))

    definition = {
        'prog':                cmdline.get_command_path(),
        'help':                cmdline.help,
        'aliases':             cmdline.aliases,
        'wraps':               cmdline.wraps,
        'abbreviate_commands': cmdline.abbreviate_commands,
        'abbreviate_options':  cmdline.abbreviate_options,
        'inherit_options':     cmdline.inherit_options,
        'options':             options,
        'positionals':         positionals,
    }

    return _mk_value_without_trace(definition)


def _commandlines_to_definition_list(commandline):
    result = []

    def callback(cmdline):
        result.append(_commandline_to_definition(cmdline))

    commandline.visit_commandlines(callback)

    return result


def validate_commandlines(cmdline):
    '''Validate commandlines.'''
    definitions = _commandlines_to_definition_list(cmdline)
    scheme_validator.validate(definitions)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""Functions for loading a python file."""

import os
import sys
import importlib
import tempfile
import __main__


def execute_file(file):
    '''Import file using compile + exec.'''
    with open(os.devnull, 'w', encoding='utf-8') as null_fh:
        sys.stdout = null_fh
        sys.stderr = null_fh

        try:
            with open(file, 'r', encoding='utf-8') as fh:
                source = fh.read()
                compiled = compile(source, file, 'exec')
                try:
                    # pylint: disable=exec-used
                    exec(compiled, globals())
                except SystemExit:
                    pass
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__

    return __main__


def import_file(file):
    '''Import file using importlib.'''
    directory, filename = os.path.split(file)
    if filename.lower().endswith('.py'):
        module_name = filename[:-3]
    else:
        temp = tempfile.NamedTemporaryFile(mode='w',
                                           encoding='utf-8',
                                           suffix='.py')

        with open(file, 'r', encoding='utf-8') as fh:
            temp.file.write(fh.read())
            temp.flush()

        directory, file = os.path.split(temp.name)
        module_name = file[:-3]

    if not directory:
        directory = '.'

    if directory not in sys.path:
        sys.path.append(directory)

    return importlib.import_module(module_name)



#!/usr/bin/env python3

import argparse

from crazy_complete import argparse_mod

argp = argparse.ArgumentParser(prog='example', description='Example program')

argp.add_argument('--version', action='version')

subp = argp.add_subparsers(description='commands')

cmdp = subp.add_parser('start',       help='Start a process').aliases(['launch'])
cmdp.add_argument('command',          help='Specify a command that shall be run').complete('command')
cmdp.add_argument('--change-dir',     help='Change to directory').complete('directory')
cmdp.add_argument('--mode', '-m',     help='Specify mode', choices=['auto', 'manual', 'debug'])

cmdp = subp.add_parser('stop',        help='Stop a process').aliases(['kill'])
cmdp.add_argument('--force', '-f',    help='Force stopping the process', action='store_true')



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Module containing code for parsing strings as a stream.'''


class StringStream:
    '''String stream class for reading a string.'''

    def __init__(self, string):
        self.s = string
        self.i = 0
        self.length = len(string)

    def peek(self, seek=0):
        '''Return character at current position + seek without advancing.'''

        try:
            return self.s[self.i + seek]
        except IndexError:
            return None

    def peek_str(self, length=1):
        '''Return substring of given length from current position.'''

        return self.s[self.i:length]

    def get(self):
        '''Return current character and advance position by one.'''

        c = self.s[self.i]
        self.i += 1
        return c

    def have(self):
        '''Return True if iterator is not at its end.'''

        return self.i < self.length

    def advance(self, length):
        '''Advance the stream by a given length.'''

        self.i += length

    def parse_shell_single_quote(self, in_quotes=False):
        '''Parse a single-quoted shell string.'''

        if not in_quotes:
            assert self.get() == "'"

        string = ''

        while True:
            try:
                c = self.get()
            except IndexError as e:
                raise ValueError("Unclosed single quote") from e

            if c == "'":
                return string

            string += c

    def parse_shell_double_quote(self, in_quotes=False):
        '''Parse a double-quoted shell string.'''

        if not in_quotes:
            assert self.get() == '"'

        string = ''

        while True:
            try:
                c = self.get()
            except IndexError as e:
                raise ValueError("Unclosed double quote") from e

            if c == '\\':
                try:
                    string += self.get()
                except IndexError as e:
                    raise ValueError("Missing character after backslash") from e

            elif c == '"':
                return string

            else:
                string += c



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for creating the subcommand switch code in Bash'''

from collections import OrderedDict

from . import utils
from . import shell
from . import bash_patterns
from .str_utils import indent


def get_subcommand_path(commandline):
    '''Return the path for `commandline`.'''
    commandlines = commandline.get_parents(include_self=True)[1:]
    prognames = ['root'] + [c.prog for c in commandlines]
    return ':'.join(prognames)


def make_subcommand_switch_code(commandline):
    '''
    Generate Bash code that tracks and updates the current subcommand path.

    This function walks through all known command lines (starting from the
    given top-level `commandline`) and generates conditional Bash code that
    appends the currently matched subcommand to the `cmd` variable. This
    allows the generated completion script to keep track of the full
    subcommand hierarchy as the user types.

    While tracking, all variations of a subcommand are considered, including
    defined aliases and any supported abbreviations. This ensures that the
    resulting `cmd` variable always contains the canonical subcommand name,
    regardless of how the user typed it.
    '''

    assert commandline.parent is None, \
        "This function should only be used on top-level command lines"

    cases = []

    for cmdline in commandline.get_all_commandlines():
        if not cmdline.get_subcommands():
            continue

        positional_num = cmdline.get_subcommands().get_positional_num()
        command_path   = get_subcommand_path(cmdline)
        switch_code    = _make_switch_case(cmdline)

        cases.append(('%s|%d' % (command_path, positional_num), switch_code))

    if not cases:
        return None

    r  = 'case "$cmd|${#POSITIONALS[@]}" in\n'
    for case in cases:
        if '\n' not in case[1]:
            r += '  %s) %s;;\n' % (shell.quote(case[0]), case[1])
        else:
            r += '  %s)\n%s;;\n' % (shell.quote(case[0]), indent(case[1], 4))
    r += 'esac'

    return r


def _make_switch_case(cmdline):
    subcommand_aliases = OrderedDict()

    for subcommand in cmdline.get_subcommands().subcommands:
        aliases = utils.get_all_command_variations(subcommand)
        aliases.remove(subcommand.prog)

        if aliases:
            subcommand_aliases[subcommand.prog] = aliases

    # We have no aliases or abbreviations, just add $arg to $cmd
    if not subcommand_aliases:
        return 'cmd+=":$arg"'

    r  = 'case "$arg" in\n'
    for command, aliases in subcommand_aliases.items():
        r += '  %s) cmd+=":%s";;\n' % (bash_patterns.make_pattern(aliases), command)
    r += '  *) cmd+=":$arg";;\n'
    r += 'esac'
    return r



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for maintaining backward compatibility with previous versions.'''

from . import utils

# Wrong types in configuration structures are silently ignored, because
# they are handled elsewhere


def fix_option_dictionary(dictionary):
    '''Fix an option dictionary.'''

    if not isinstance(dictionary, dict):
        return

    if 'group' in dictionary:
        if 'groups' in dictionary:
            msg = ('Both `group` and `groups` found. `group` is deprecated. '
                   'Removing `group` in favour of `groups`')
            utils.warn(msg)
            dictionary.pop('group')
        else:
            utils.warn('`group` is deprecated. Please use `groups` instead')
            dictionary['groups'] = [dictionary.pop('group')]


def fix_commandline_dictionary(dictionary):
    '''Fix a commandline dictionary.'''

    if not isinstance(dictionary, dict):
        return

    options = dictionary.get('options', [])

    if isinstance(options, list):
        for option in options:
            fix_option_dictionary(option)


def fix_commandline_dictionaries(dictionaries):
    '''Fix a list of commandline dictionaries.'''

    if not isinstance(dictionaries, list):
        return

    for commandline_dictionary in dictionaries:
        fix_commandline_dictionary(commandline_dictionary)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Value holder class with line and column information'''


class ValueWithTrace:
    '''Represents a value from a configuration file along with its metadata.

    This class is used to associate a value from a YAML or JSON file with its
    source, line number, and column number. It is particularly useful for
    debugging and error reporting, as it provides precise location information
    for the value in the original configuration file.

    Attributes:
        value (Any): The value extracted from the configuration file.
        src (str): The source of the configuration file.
        line (int): The line number of the value in the source.
        column (int): The column number of the value in the source.
    '''

    def __init__(self, value, source, line, column):
        self.value = value
        self.source = source
        self.line = line
        self.column = column

    @staticmethod
    def from_yaml_event(value, source, event):
        '''Constructs a `ValueWithTrace` from a YAML event object.'''

        return ValueWithTrace(
            value,
            source,
            event.start_mark.line + 1,
            event.start_mark.column + 1)

    def get_position_string(self):
        '''Return a string describing the position of the value.'''

        return f'line {self.line}, column {self.column}'

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value == other

    def __repr__(self):
        return f'{self.value!r}'


class ValueWithOutTrace:
    '''Class holding a value without metadata. '''

    def __init__(self, value):
        self.value = value

    def get_position_string(self):
        '''Return empty string.'''

        return ''

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value == other

    def __repr__(self):
        return f'{self.value!r}'



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Module for creating Bash globs.'''

from collections import defaultdict

from .algo import numbers_are_contiguous


class TrieNode:
    '''Trie class.'''

    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.children = defaultdict(TrieNode)
        self.is_terminal = False

    def insert(self, string):
        '''Insert string into trie.'''

        node = self
        for char in string:
            node = node.children[char]
        node.is_terminal = True


def _build_pattern(parts):
    if len(parts) == 1:
        return parts[0]

    try:
        ints = sorted(map(int, parts))
        if numbers_are_contiguous(ints):
            return f'[{ints[0]}-{ints[-1]}]'
    except ValueError:
        pass

    return "@(" + "|".join(sorted(parts)) + ")"


def trie_to_pattern(node):
    '''
    Recursively convert the trie into a Bash extglob pattern.
    '''
    if not node.children:
        return ""

    parts = []
    for token, child in node.children.items():
        sub = trie_to_pattern(child)
        if sub:
            parts.append(token + sub)
        else:
            parts.append(token)

    # Terminal means that this path can end here
    if node.is_terminal:
        parts.append("")  # Represent the option to stop early

    return _build_pattern(parts)


def compact_glob_trie(strings):
    '''
    Build a recursive glob pattern that matches all input strings.
    '''
    strings = sorted(set(strings))
    if len(strings) == 1:
        return strings[0]

    # Build Trie
    root = TrieNode()
    for string in strings:
        root.insert(string)

    # Convert Trie to pattern
    return trie_to_pattern(root)


def make_pattern(strings):
    '''Make a glob pattern that matches `strings`.'''

    candidate0 = '|'.join(strings)
    candidate1 = compact_glob_trie(strings)

    return candidate0 if len(candidate0) < len(candidate1) else candidate1



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for validating the structure of a command line definition.'''

from types import NoneType

from .cli import ExtendedBool, is_extended_bool
from .when import parse_when
from .errors import CrazyError, CrazySchemaValidationError
from .pattern import bash_glob_to_regex, bash_glob_to_zsh_glob
from .value_with_trace import ValueWithTrace
from .str_utils import (
    contains_space, is_empty_or_whitespace,
    is_valid_option_string, is_valid_variable_name,
    is_valid_extended_regex, validate_prog)
from . import messages as m


_error = CrazySchemaValidationError


# =============================================================================
# Helper functions for validating
# =============================================================================


class Context:
    '''Class for providing context to validation.'''

    # pylint: disable=too-few-public-methods

    TYPE_OPTION = 0
    TYPE_POSITIONAL = 1

    def __init__(self):
        self.definition = None
        self.option = None
        self.positional = None
        self.type = None
        self.trace = None


class Arguments:
    '''Class for accessing arguments.'''

    def __init__(self, args):
        self.args = args
        self.index = 0

    def get_required_arg(self, name):
        '''Return a required argument, else raise an exception.'''

        if self.index < len(self.args.value):
            arg = self.args.value[self.index]
            self.index += 1
            return arg

        raise _error('%s: %s' % (m.missing_arg(), name), self.args)

    def get_optional_arg(self, default=None):
        '''Return an optional arg, else return a default.'''

        if self.index < len(self.args.value):
            arg = self.args.value[self.index]
            self.index += 1
            return arg

        return ValueWithTrace(default, '<default value>', 1, 1)

    def require_no_more(self):
        '''Raise an exception if there are any arguments left.'''

        if self.index < len(self.args.value):
            raise _error(m.too_many_arguments(), self.args)


def _has_set(dictionary, key):
    return (key in dictionary.value and
            dictionary.value[key].value is not None)


def _is_true(dictionary, key):
    return (key in dictionary.value and
            dictionary.value[key].value is True)


def _check_type(value, types, parameter_name=None):
    if not isinstance(value.value, types):
        types_strings = []
        for t in types:
            types_strings.append({
                str:      'string',
                int:      'integer',
                float:    'float',
                list:     'list',
                dict:     'dictionary',
                bool:     'boolean',
                NoneType: 'none'}[t])
        types_string = '|'.join(types_strings)

        msg = m.invalid_type_expected_types(types_string)
        if parameter_name is not None:
            msg = f'{parameter_name}: {msg}'
        raise _error(msg, value)


def _check_dictionary(dictionary, rules):
    _check_type(dictionary, (dict,))

    for key, value in dictionary.value.items():
        if key.value not in rules:
            raise _error('%s: %s' % (m.unknown_parameter(), key.value), key)

        _check_type(value, rules[key.value][1], key.value)

        if rules[key.value][2]:
            rules[key.value][2](value, key.value)

    for key, rule in rules.items():
        if rule[0] is True and key not in dictionary.value:
            raise _error('%s: %s' % (m.missing_arg(), key), dictionary)


def _check_regex(value, parameter):
    if not is_valid_extended_regex(value.value):
        raise _error('%s: %s' % (parameter, m.not_an_extended_regex()), value)


def _check_char(value, parameter):
    if len(value.value) != 1:
        msg = '%s: %s' % (parameter, m.single_character_expected())
        raise _error(msg, value)


def _check_variable_name(value, parameter):
    if not is_valid_variable_name(value.value):
        raise _error('%s: %s' % (parameter, m.not_a_variable_name()), value)


def _check_non_empty_string(value, parameter):
    if is_empty_or_whitespace(value.value):
        raise _error('%s: %s' % (parameter, m.string_cannot_be_empty()), value)


def _check_no_spaces(value, parameter):
    if contains_space(value.value):
        msg = '%s: %s' % (parameter, m.string_cannot_contain_space())
        raise _error(msg, value)


def _check_non_empty_list(value, parameter):
    if len(value.value) == 0:
        raise _error('%s: %s' % (parameter, m.list_cannot_be_empty()), value)


def _check_non_empty_dict(value, parameter):
    if len(value.value) == 0:
        raise _error('%s: %s' % (parameter, m.dict_cannot_be_empty()), value)


def _check_extended_bool(value, parameter):
    if not is_extended_bool(value.value):
        expected = 'true, false or `%s`' % ExtendedBool.INHERIT
        msg = '%s: %s' % (parameter, m.invalid_value_expected_values(expected))
        raise _error(msg, value)


# =============================================================================
# Actual validation code
# =============================================================================


def _check_when(value):
    try:
        parse_when(value.value)
    except CrazyError as e:
        raise _error(f'when: {e}', value) from e


def _check_void(_ctxt, arguments):
    arguments.require_no_more()


def _check_none(ctxt, arguments):
    arguments.require_no_more()

    if ctxt.trace and ctxt.trace[-1] in ('combine', 'list', 'prefix'):
        msg = m.completer_not_allowed_in('none', ctxt.trace[-1])
        raise _error(msg, arguments.args)


def _check_choices(_ctxt, arguments):
    choices = arguments.get_required_arg('values')
    _check_type(choices, (list, dict), 'values')
    arguments.require_no_more()

    if isinstance(choices.value, dict):
        for i, (value, desc) in enumerate(choices.value.items()):
            _check_type(value, (str, int), f'values[{i}]: item')
            _check_type(desc,  (str, int), f'values[{i}]: description')

    elif isinstance(choices.value, list):
        for i, value in enumerate(choices.value):
            _check_type(value, (str, int), f'values[{i}]')


def _check_command(_ctxt, arguments):
    opts = arguments.get_optional_arg({})
    _check_type(opts, (dict,), 'options')
    arguments.require_no_more()

    _check_dictionary(opts, {
        'path':         (False, (str,), _check_non_empty_string),
        'path_append':  (False, (str,), _check_non_empty_string),
        'path_prepend': (False, (str,), _check_non_empty_string),
    })

    path = opts.value.get('path', None)
    append = opts.value.get('path_append', None)
    prepend = opts.value.get('path_prepend', None)

    if path and append:
        msg = m.mutually_exclusive_parameters('path, path_append')
        raise _error(msg, opts)

    if path and prepend:
        msg = m.mutually_exclusive_parameters('path, path_prepend')
        raise _error(msg, opts)


def _check_command_arg(ctxt, arguments):
    arguments.require_no_more()

    if ctxt.trace and ctxt.trace[-1] in ('combine', 'list', 'key_value_list'):
        msg = m.completer_not_allowed_in('command_arg', ctxt.trace[-1])
        raise _error(msg, arguments.args)

    if ctxt.type != Context.TYPE_POSITIONAL:
        msg = m.completer_not_allowed_in_option('command_arg')
        raise _error(msg, arguments.args)

    if (not _has_set(ctxt.positional, 'repeatable') or
        not ctxt.positional.value['repeatable'].value):
        msg = m.completer_requires_repeatable('command_arg')
        raise _error(msg, ctxt.positional)

    def command_is_previous_to_command_arg(positional):
        return (
            _has_set(positional, 'complete') and
            positional.value['complete'].value[0] == 'command' and
            positional.value['number'].value + 1 == ctxt.positional.value['number'].value
        )

    positionals = ctxt.definition.value['positionals'].value
    if not any(filter(command_is_previous_to_command_arg, positionals)):
        msg = m.command_arg_without_command()
        raise _error(msg, ctxt.positional)


def _check_filedir(_ctxt, arguments, with_file_opts=False, with_list_opts=False):
    options = arguments.get_optional_arg({})
    _check_type(options, (dict,), 'options')
    arguments.require_no_more()

    spec = {'directory': (False, (str,), _check_non_empty_string)}

    if with_file_opts:
        spec['extensions'] = (False, (list,), _check_non_empty_list)
        spec['fuzzy'] = (False, (bool,), None)
        spec['ignore_globs'] = (False, (list,), _check_non_empty_list)

    if with_list_opts:
        spec['separator'] = (False, (str,), _check_char)
        spec['duplicates'] = (False, (bool,), None)

    _check_dictionary(options, spec)

    if _has_set(options, 'extensions'):
        for i, extension in enumerate(options.value['extensions'].value):
            _check_type(extension, (str,), f'extensions[{i}]')
            _check_non_empty_string(extension, f'extensions[{i}]')
            _check_no_spaces(extension, f'extensions[{i}]')

    if _has_set(options, 'ignore_globs'):
        for pattern in options.value['ignore_globs'].value:
            _check_type(pattern, (str,), 'pattern')
            _check_non_empty_string(pattern, 'pattern')
            try:
                bash_glob_to_regex(pattern.value)
                bash_glob_to_zsh_glob(pattern.value)
            except ValueError as e:
                raise _error('%s: %s' % ('pattern', e), pattern) from e


def _check_file(ctxt, arguments):
    _check_filedir(ctxt, arguments, with_file_opts=True)


def _check_directory(ctxt, arguments):
    _check_filedir(ctxt, arguments)


def _check_file_list(ctxt, arguments):
    _check_filedir(ctxt, arguments, with_file_opts=True, with_list_opts=True)


def _check_directory_list(ctxt, arguments):
    _check_filedir(ctxt, arguments, with_list_opts=True)


def _check_mime_file(_ctxt, arguments):
    pattern = arguments.get_required_arg("pattern")
    _check_type(pattern, (str,), "pattern")
    arguments.require_no_more()
    _check_regex(pattern, 'pattern')


def _check_range(_ctxt, arguments):
    start = arguments.get_required_arg("start")
    stop = arguments.get_required_arg("stop")
    step = arguments.get_optional_arg(1)
    arguments.require_no_more()

    _check_type(start, (int,), "start")
    _check_type(stop, (int,), "stop")
    _check_type(step, (int,), "step")

    if step.value == 0:
        msg = '%s: %s' % ('step', m.integer_cannot_be_zero())
        raise _error(msg, step)

    if step.value > 0 and start.value > stop.value:
        msg = f"start > stop: {start.value} > {stop.value} (step={step.value})"
        raise _error(msg, step)

    if step.value < 0 and stop.value > start.value:
        msg = f"stop > start: {stop.value} > {start.value} (step={step.value})"
        raise _error(msg, step)


def _check_exec(_ctxt, arguments):
    command = arguments.get_required_arg("command")
    _check_type(command, (str,), "command")
    arguments.require_no_more()


def _check_value_list(_ctxt, arguments):
    options = arguments.get_required_arg('options')
    _check_type(options, (dict,), 'options')
    arguments.require_no_more()

    _check_dictionary(options, {
        'values':    (True,  (list, dict), None),
        'separator': (False, (str,), _check_char),
        'duplicates': (False, (bool,), None),
    })

    values = options.value['values']

    if isinstance(values.value, dict):
        _check_non_empty_dict(values, 'values')

        for i, (item, desc) in enumerate(values.value.items()):
            _check_type(item, (str,), f'values[{i}]: item')
            _check_type(desc, (str,), f'values[{i}]: description')
    else:
        _check_non_empty_list(values, 'values')

        for i, value in enumerate(values.value):
            _check_type(value, (str,), f'values[{i}]: item')


def _check_key_value_list(ctxt, arguments):
    pair_separator = arguments.get_required_arg('pair_separator')
    value_separator = arguments.get_required_arg('value_separator')
    values = arguments.get_required_arg('values')
    arguments.require_no_more()

    _check_type(pair_separator, (str,), 'pair_separator')
    _check_type(value_separator, (str,), 'value_separator')
    _check_type(values, (list,), 'values')

    ctxt.trace.append('key_value_list')

    _check_char(pair_separator, 'pair_separator')
    _check_char(value_separator, 'value_separator')
    _check_non_empty_list(values, 'values')

    for i, compldef in enumerate(values.value):
        index = f'definition[{i}]'
        _check_type(compldef, (list,), index)

        if len(compldef.value) != 3:
            msg = '%s: %s' % (index, m.list_must_contain_exact_three_items())
            raise _error(msg, compldef)

        key = compldef.value[0]
        description = compldef.value[1]
        complete = compldef.value[2]

        _check_type(key, (str,), index + ': key')
        _check_type(description, (str, NoneType), index + ': description')
        _check_type(complete, (list, NoneType), index + ': completer')
        _check_non_empty_string(key, index + ': key')
        _check_no_spaces(key, index + ': key')

        if complete.value is None:
            continue

        _check_complete(ctxt, complete)


def _check_combine(ctxt, arguments):
    commands = arguments.get_required_arg('commands')
    arguments.require_no_more()

    _check_type(commands, (list,), 'commands')
    _check_non_empty_list(commands, 'commands')

    if len(commands.value) == 1:
        msg = '%s: %s' % ('commands', m.list_must_contain_at_least_two_items())
        raise _error(msg, commands)

    ctxt.trace.append('combine')

    for i, completer in enumerate(commands.value):
        _check_type(completer, (list,), f'command[{i}]')
        _check_complete(ctxt, completer)


def _check_list(ctxt, arguments):
    command = arguments.get_required_arg('command')
    options = arguments.get_optional_arg({})
    arguments.require_no_more()

    _check_type(command, (list,), 'command')
    _check_type(options, (dict,), 'options')

    _check_dictionary(options, {
        'separator': (False, (str,), _check_char),
        'duplicates': (False, (bool,), None),
    })

    ctxt.trace.append('list')
    _check_complete(ctxt, command)


def _check_prefix(ctxt, arguments):
    prefix = arguments.get_required_arg('prefix')
    command = arguments.get_required_arg('command')
    arguments.require_no_more()

    _check_type(prefix, (str,), 'prefix')
    _check_type(command, (list,), 'command')

    ctxt.trace.append('prefix')
    _check_complete(ctxt, command)


def _check_history(_ctxt, arguments):
    pattern = arguments.get_required_arg("pattern")
    _check_type(pattern, (str,), "pattern")
    arguments.require_no_more()
    _check_regex(pattern, 'pattern')


def _check_date(_ctxt, arguments):
    format_ = arguments.get_required_arg("format")
    _check_type(format_, (str,), "format")
    arguments.require_no_more()
    _check_non_empty_string(format_, 'format')


def _check_ip_address(_ctxt, arguments):
    type_ = arguments.get_optional_arg('all')
    arguments.require_no_more()
    _check_type(type_, (str,), "type")

    if type_.value not in ('ipv4', 'ipv6', 'all'):
        allowed = 'ipv4, ipv6, all'
        msg = '%s: %s' % ('type', m.invalid_value_expected_values(allowed))
        raise _error(msg, type_)


def _check_integer_float(_ctxt, arguments, types):
    options = arguments.get_optional_arg({})
    arguments.require_no_more()
    _check_type(options, (dict,), "options")

    _check_dictionary(options, {
        'min':      (False, types, None),
        'max':      (False, types, None),
        'help':     (False, (str,), _check_non_empty_string),
        'suffixes': (False, (dict,), _check_non_empty_dict),
    })

    if _has_set(options, 'min') and _has_set(options, 'max'):
        min_ = options.value['min'].value
        max_ = options.value['max'].value

        if min_ > max_:
            raise _error(f'min > max ({min_} > {max_})', options)

    if _has_set(options, 'suffixes'):
        for suffix, description in options.value['suffixes'].value.items():
            _check_type(suffix, (str,), "suffix")
            _check_type(description, (str,), "description")
            _check_non_empty_string(suffix, "suffix")
            _check_non_empty_string(description, "description")


def _check_integer(ctxt, arguments):
    _check_integer_float(ctxt, arguments, (int,))


def _check_float(ctxt, arguments):
    _check_integer_float(ctxt, arguments, (int, float))


_COMMANDS = {
    'alsa_card':          _check_void,
    'alsa_device':        _check_void,
    'charset':            _check_void,
    'choices':            _check_choices,
    'combine':            _check_combine,
    'command':            _check_command,
    'command_arg':        _check_command_arg,
    'commandline_string': _check_void,
    'date':               _check_date,
    'date_format':        _check_void,
    'directory':          _check_directory,
    'directory_list':     _check_directory_list,
    'environment':        _check_void,
    'exec':               _check_exec,
    'exec_fast':          _check_exec,
    'exec_internal':      _check_exec,
    'file':               _check_file,
    'file_list':          _check_file_list,
    'filesystem_type':    _check_void,
    'float':              _check_float,
    'gid':                _check_void,
    'group':              _check_void,
    'history':            _check_history,
    'hostname':           _check_void,
    'integer':            _check_integer,
    'ip_address':         _check_ip_address,
    'key_value_list':     _check_key_value_list,
    'list':               _check_list,
    'locale':             _check_void,
    'login_shell':        _check_void,
    'mime_file':          _check_mime_file,
    'mountpoint':         _check_void,
    'net_interface':      _check_void,
    'none':               _check_none,
    'pid':                _check_void,
    'prefix':             _check_prefix,
    'process':            _check_void,
    'range':              _check_range,
    'service':            _check_void,
    'signal':             _check_void,
    'timezone':           _check_void,
    'uid':                _check_void,
    'user':               _check_void,
    'value_list':         _check_value_list,
    'variable':           _check_void,
}


def _check_complete(ctxt, args):
    arguments = Arguments(args)
    cmd = arguments.get_required_arg('command')

    _check_type(cmd, (str,), 'command')

    if cmd.value not in _COMMANDS:
        msg = '%s: %s' % (m.unknown_completer(), cmd.value)
        raise _error(msg, cmd)

    try:
        _COMMANDS[cmd.value](ctxt, arguments)
    except _error as e:
        msg = '%s: %s' % (cmd.value, e.message)
        raise _error(msg, e.value_with_trace) from e


def _check_positionals_repeatable0(definition_tree, definition):
    if not _has_set(definition, 'positionals'):
        return

    positionals = definition.value['positionals'].value
    repeatable_number = None

    for positional in sorted(positionals, key=lambda p: p.value['number'].value):
        repeatable = False
        if _has_set(positional, 'repeatable'):
            repeatable = positional.value['repeatable'].value

        positional_number = positional.value['number'].value

        if repeatable:
            if repeatable_number is not None and repeatable_number != positional_number:
                msg = m.too_many_repeatable_positionals()
                raise _error(msg, positional)

            repeatable_number = positional_number
        elif repeatable_number is not None and positional_number > repeatable_number:
            msg = m.positional_argument_after_repeatable()
            raise _error(msg, positional)

    node = definition_tree.get_definition(definition.value['prog'].value)
    if repeatable_number is not None and len(node.subcommands) > 0:
        msg = m.repeatable_with_subcommands()
        raise _error(msg, definition)


def _check_positionals_repeatable(definition_tree, definition):
    try:
        _check_positionals_repeatable0(definition_tree, definition)
    except _error as e:
        prog = definition.value['prog'].value
        msg = '%s: %s' % (prog, e.message)
        raise _error(msg, e.value_with_trace) from e


def _check_option0(ctxt, option):
    chkbool = _check_extended_bool

    _check_dictionary(option, {
        'option_strings': (True,  (list,),               None),
        'metavar':        (False, (str,  NoneType),      None),
        'help':           (False, (str,  NoneType),      None),
        'optional_arg':   (False, (bool, NoneType),      None),
        'group':          (False, (str,  NoneType),      None),
        'groups':         (False, (list, NoneType),      None),
        'repeatable':     (False, (bool, str, NoneType), chkbool),
        'final':          (False, (bool, NoneType),      None),
        'hidden':         (False, (bool, NoneType),      None),
        'nosort':         (False, (bool, NoneType),      None),
        'complete':       (False, (list, NoneType),      None),
        'when':           (False, (str,  NoneType),      None),
        'capture':        (False, (str,  NoneType),      None),
        'long_opt_arg_sep': (False, (str,),              None),
    })

    option_strings = option.value['option_strings']

    _check_non_empty_list(option_strings, 'option_strings')

    for option_string in option_strings.value:
        _check_type(option_string, (str,), "option_string")

        if not is_valid_option_string(option_string.value):
            msg = m.invalid_value()
            raise _error(msg, option_string)

    if _has_set(option, 'metavar') and not _has_set(option, 'complete'):
        msg = m.parameter_requires_parameter('metavar', 'complete')
        raise _error(msg, option)

    if _is_true(option, 'optional_arg') and not _has_set(option, 'complete'):
        msg = m.parameter_requires_parameter('optional_arg', 'complete')
        raise _error(msg, option)

    if _is_true(option, 'repeatable') and _is_true(option, 'hidden'):
        msg = m.mutually_exclusive_parameters('repeatable, hidden')
        raise _error(msg, option)

    if _has_set(option, 'group') and _has_set(option, 'groups'):
        msg = m.mutually_exclusive_parameters('groups, group')
        raise _error(msg, option)

    if _has_set(option, 'groups'):
        for group in option.value['groups'].value:
            _check_type(group, (str,), "group")

    if _has_set(option, 'complete'):
        ctxt.type = Context.TYPE_OPTION
        ctxt.option = option
        ctxt.trace = []
        _check_complete(ctxt, option.value['complete'])

    if _has_set(option, 'when'):
        _check_when(option.value['when'])

    if _has_set(option, 'capture'):
        _check_variable_name(option.value['capture'], 'capture')

    if _has_set(option, 'long_opt_arg_sep'):
        if option.value['long_opt_arg_sep'] not in (
                'space', 'equals', 'both', ExtendedBool.INHERIT):
            expected = f'space, equals, both, {ExtendedBool.INHERIT}'
            msg = m.invalid_value_expected_values(expected)
            raise _error(msg, option.value['long_opt_arg_sep'])


def _check_option(ctxt, option):
    try:
        _check_option0(ctxt, option)
    except _error as e:
        try:
            option_strings = option.value['option_strings'].value
            option_strings = '|'.join([o.value for o in option_strings])
        except (KeyError, TypeError):
            raise e from None

        msg = '%s: %s' % (option_strings, e.message)
        raise _error(msg, e.value_with_trace) from e


def _check_positional0(ctxt, positional):
    _check_dictionary(positional, {
        'number':     (True,  (int,),           None),
        'metavar':    (False, (str,  NoneType), None),
        'help':       (False, (str,  NoneType), None),
        'repeatable': (False, (bool, NoneType), None),
        'nosort':     (False, (bool, NoneType), None),
        'complete':   (False, (list, NoneType), None),
        'when':       (False, (str,  NoneType), None),
        'capture':    (False, (str,  NoneType), None),
    })

    if positional.value['number'].value < 1:
        msg = m.integer_must_be_greater_than_zero()
        raise _error('%s: %s' % ('number', msg), positional.value['number'])

    if _has_set(positional, 'complete'):
        ctxt.type = Context.TYPE_POSITIONAL
        ctxt.positional = positional
        ctxt.trace = []
        _check_complete(ctxt, positional.value['complete'])

    if _has_set(positional, 'when'):
        _check_when(positional.value['when'])

    if _has_set(positional, 'capture'):
        _check_variable_name(positional.value['capture'], 'capture')


def _check_positional(ctxt, positional):
    try:
        _check_positional0(ctxt, positional)
    except _error as e:
        metavar = 'unknown'

        try:
            metavar = 'positional #%d' % positional.value['number'].value
        except (KeyError, TypeError):
            pass

        try:
            metavar_ = positional.value['metavar'].value
            assert isinstance(metavar_, str)
            metavar = metavar_
        except (AssertionError, KeyError):
            pass

        msg = '%s: %s' % (metavar, e.message)
        raise _error(msg, e.value_with_trace) from e


def _check_commandline_definition0(ctxt, definition):
    chkbool = _check_extended_bool

    _check_dictionary(definition, {
        'prog':                (True,  (str,),                None),
        'help':                (False, (str,  NoneType),      None),
        'aliases':             (False, (list, NoneType),      None),
        'wraps':               (False, (str,  NoneType),      None),
        'abbreviate_commands': (False, (bool, str, NoneType), chkbool),
        'abbreviate_options':  (False, (bool, str, NoneType), chkbool),
        'inherit_options':     (False, (bool, str, NoneType), chkbool),
        'options':             (False, (list, NoneType),      None),
        'positionals':         (False, (list, NoneType),      None),
    })

    try:
        validate_prog(definition.value['prog'].value)
    except CrazyError as e:
        raise _error(f'prog: {e}', definition.value['prog']) from None

    if _has_set(definition, 'aliases'):
        for alias in definition.value['aliases'].value:
            _check_type(alias, (str,), 'alias')
            _check_non_empty_string(alias, 'alias')
            _check_no_spaces(alias, 'alias')

    if _has_set(definition, 'wraps'):
        wraps = definition.value['wraps']
        _check_non_empty_string(wraps, 'wraps')
        _check_no_spaces(wraps, 'wraps')

        if contains_space(definition.value['prog'].value):
            msg = m.parameter_not_allowed_in_subcommand('wraps')
            raise _error(msg, definition.value['wraps'])

    if _has_set(definition, 'options'):
        for option in definition.value['options'].value:
            _check_option(ctxt, option)

    if _has_set(definition, 'positionals'):
        for positional in definition.value['positionals'].value:
            _check_positional(ctxt, positional)


def _check_commandline_definition(ctxt, definition):
    try:
        _check_commandline_definition0(ctxt, definition)
    except _error as e:
        try:
            prog = definition.value['prog'].value
        except (KeyError, TypeError):
            prog = 'unknown'

        msg = '%s: %s' % (prog, e.message)
        raise _error(msg, e.value_with_trace) from e


class DefinitionTree:
    '''Holds a command-line definition with subcommands.'''

    def __init__(self, definition):
        self.definition = definition
        self.subcommands = {}

    def add_definition(self, definition):
        '''Add a definition to subcommands.'''

        _check_type(definition, (dict,))

        if not _has_set(definition, 'prog'):
            raise _error('%s: %s' % (m.missing_arg(), 'prog'), definition)

        try:
            validate_prog(definition.value['prog'].value)
        except CrazyError as e:
            raise _error(f'prog: {e}', definition.value['prog']) from None

        commands = definition.value['prog'].value.split(' ')
        subcommand = commands.pop(-1)

        node = self

        for i, part in enumerate(commands):
            try:
                node = node.subcommands[part]
            except KeyError:
                prog = ' '.join(commands[0:i+1])
                msg = m.missing_definition_of_program(prog)
                raise _error(msg, definition) from None

        if subcommand in node.subcommands:
            prog = definition.value['prog'].value
            msg = m.multiple_definition_of_program(prog)
            raise _error(msg, definition)

        node.subcommands[subcommand] = DefinitionTree(definition)

    def get_definition(self, prog):
        '''Get definition by prog.'''

        commands = prog.split(' ')
        node = self

        for part in commands:
            node = node.subcommands[part]

        return node

    def visit(self, callback):
        '''Call `callback` on each node in the tree.'''
        callback(self)
        for subcommand in self.subcommands.values():
            callback(subcommand)

    def get_all(self):
        '''Return a list of all nodes.'''
        result = []
        self.visit(result.append)
        return result

    def get_first_commandline(self):
        '''Return the first command line in the tree.'''
        return list(self.subcommands.values())[0]

    @staticmethod
    def make_tree(definition_list):
        '''Make a tree out of a list of definitions.'''

        root = DefinitionTree(None)

        for definition in definition_list:
            root.add_definition(definition)

        return root


def validate(definition_list):
    '''Validate a list of definitions.'''

    context = Context()
    commandline_root = DefinitionTree.make_tree(definition_list)

    if len(commandline_root.subcommands) == 0:
        msg = m.no_programs_defined()
        raise _error(msg, ValueWithTrace(None, '', 1, 1))

    if len(commandline_root.subcommands) > 1:
        value = ValueWithTrace(None, '', 1, 1)
        progs = list(commandline_root.subcommands.keys())
        msg = m.too_many_programs_defined()
        raise _error('%s: %s' % (msg, progs), value)

    for cmdline in commandline_root.get_first_commandline().get_all():
        context.definition = cmdline.definition
        _check_commandline_definition(context, cmdline.definition)
        _check_positionals_repeatable(commandline_root, cmdline.definition)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''YAML Parser with column and line information.'''

import yaml
from yaml.events import (StreamStartEvent,
                         DocumentStartEvent, DocumentEndEvent,
                         MappingStartEvent, MappingEndEvent,
                         SequenceStartEvent, SequenceEndEvent,
                         ScalarEvent, AliasEvent)

from .value_with_trace import ValueWithTrace
from .errors import CrazySchemaValidationError


_error = CrazySchemaValidationError


class ExtendedYAMLParser:
    '''Parse YAML with the ability to trace the origin of parsed values.'''

    def __init__(self):
        self.src = None
        self.data = []
        self.current_stack = []
        self.current_key = None

    def parse(self, stream):
        """
        Parses the given YAML stream in a SAX-like way and reconstructs the
        structure.

        Raises:
            - yaml.parser.ParserError
            - errors.CrazySchemaValidationError
        """
        self.src = stream
        self.data = []
        self.current_stack = []
        self.current_key = None

        loader = yaml.Loader(stream)
        try:
            while True:
                event = loader.get_event()
                if event is None:
                    break
                self.handle_event(event)
        finally:
            loader.dispose()

        return self.data

    def handle_event(self, event):
        """
        Handle each YAML parsing event and construct the corresponding data
        structure.
        """
        if isinstance(event, StreamStartEvent):
            pass
        elif isinstance(event, DocumentStartEvent):
            pass
        elif isinstance(event, DocumentEndEvent):
            if self.current_stack:
                self.data.append(self.current_stack.pop())
        elif isinstance(event, MappingStartEvent):
            new_map = ValueWithTrace.from_yaml_event({}, self.src, event)
            self.add_to_current_structure(self.current_key, new_map)
            self.current_stack.append(new_map)
            self.current_key = None
        elif isinstance(event, SequenceStartEvent):
            new_seq = ValueWithTrace.from_yaml_event([], self.src, event)
            self.add_to_current_structure(self.current_key, new_seq)
            self.current_stack.append(new_seq)
            self.current_key = None
        elif isinstance(event, ScalarEvent):
            value = self.convert_scalar(event.value, event.implicit)
            value = ValueWithTrace.from_yaml_event(value, self.src, event)

            if self.current_stack and isinstance(self.current_stack[-1].value, list):
                self.add_to_current_structure(None, value)
            else:
                if self.current_key is None:
                    self.current_key = value
                else:
                    self.add_to_current_structure(self.current_key, value)
                    self.current_key = None
        elif isinstance(event, (MappingEndEvent, SequenceEndEvent)):
            self.current_stack.pop()
        elif isinstance(event, AliasEvent):
            value = ValueWithTrace.from_yaml_event(None, self.src, event)
            raise _error('Anchors not supported', value)

    def add_to_current_structure(self, key, value):
        """
        Add a value to the current structure (either dict or list).
        """
        if not self.current_stack:
            # Root-level element
            self.current_stack.append(value)
            return

        current = self.current_stack[-1]
        if isinstance(current.value, dict):
            if key is None:
                raise _error("Missing key for mapping in YAML", value)
            current.value[key] = value
        elif isinstance(current.value, list):
            current.value.append(value)
        else:
            raise _error("Invalid structure in YAML", value)

    def convert_scalar(self, value, implicit):
        """
        Converts a scalar based on its type hints (implicit tuple).
        """
        if implicit[0]:
            lower = value.lower()

            if lower in ("null", "~", ""):
                return None

            if lower in ("true", "on", "yes"):
                return True

            if lower in ("false", "off", "no"):
                return False

            try:
                return int(value)
            except ValueError:
                pass

            try:
                return float(value)
            except ValueError:
                pass

        return value



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for generating a Zsh auto completion file.'''

from collections import namedtuple, OrderedDict

from . import config as config_
from . import generation
from . import shell
from . import utils
from . import zsh_complete
from . import zsh_helpers
from . import zsh_utils
from . import zsh_when
from . import zsh_wrapper
from .output import Output
from .str_utils import indent


Arg = namedtuple('Arg', ('option', 'when', 'option_spec'))


class ZshQuery:
    '''Helper class for using the `query` function.'''

    # pylint: disable=too-few-public-methods

    def __init__(self, ctxt):
        self.ctxt = ctxt
        self.used = False

    def use(self, command=None):
        '''Return the function name of query and use a command.'''

        self.used = True
        return self.ctxt.helpers.use_function('query', command)


class ZshCompletionFunction:
    '''Class for generating a zsh completion function.'''

    # pylint: disable=too-many-branches
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    def __init__(self, ctxt, commandline):
        self.commandline = commandline
        self.ctxt = ctxt
        self.funcname = ctxt.helpers.make_completion_funcname(commandline)
        self.subcommands = commandline.get_subcommands()
        self.command_counter = 0
        self.completer = zsh_complete.ZshCompleter()
        self.code = None
        self.query = ZshQuery(ctxt)
        self._generate_completion_code()

    def _complete(self, option, command, *args):
        context = self.ctxt.get_option_context(self.commandline, option)
        return self.completer.complete(context, [], command, *args)

    def _complete_option(self, option):
        if option.complete:
            action = self._complete(option, *option.complete).get_action_string()
        else:
            action = None

        option_spec = zsh_utils.make_option_spec(
            option.option_strings,
            conflicting_options = option.get_conflicting_option_strings(),
            description = option.help,
            complete = option.complete,
            optional_arg = option.optional_arg,
            repeatable = option.repeatable,
            hidden = option.hidden,
            final = option.final,
            metavar = option.metavar,
            action = action,
            long_opt_arg_sep = option.long_opt_arg_sep
        )

        return Arg(option, option.when, option_spec)

    def _complete_subcommands(self, positional):
        choices = positional.get_choices()
        self.command_counter += 1

        spec = zsh_utils.make_positional_spec(
            positional.get_positional_num(),
            False,
            f'command{self.command_counter}',
            self._complete(positional, 'choices', choices).get_action_string()
        )

        return Arg(positional, None, spec)

    def _complete_positional(self, positional):
        spec = zsh_utils.make_positional_spec(
            positional.get_positional_num(),
            positional.repeatable,
            positional.help or positional.metavar or ' ',
            self._complete(positional, *positional.complete).get_action_string()
        )

        return Arg(positional, positional.when, spec)

    def _generate_completion_code(self):
        self.code = OrderedDict()
        self.code['0-init']        = ''
        self.code['1-capture']     = ''
        self.code['2-subcommands'] = ''
        self.code['3-options']     = ''

        # We have to call these functions first, because they tell us if
        # the query function is used.
        self.code['1-capture']     = self._generate_option_capture()
        self.code['2-subcommands'] = self._generate_subcommand_call()
        self.code['3-options']     = self._generate_option_parsing()

        if self.query.used:
            r  = 'local opts=%s\n' % shell.quote(utils.get_query_option_strings(self.commandline))
            r += "local HAVING_OPTIONS=() OPTION_VALUES=() POSITIONALS=() INCOMPLETE_OPTION=''\n"
            r += '%s init "$opts" "${words[@]}"' % self.query.use()
            self.code['0-init'] = r

    def _generate_option_capture(self):
        local_vars = []
        set_cmds   = []

        for option in self.commandline.options:
            if option.capture:
                local_vars.append(option.capture)
                set_cmds.append('IFS=$"\\n" %s=($(%s get_option %s))' % (
                    option.capture,
                    self.query.use('get_option'),
                    shell.join_quoted(option.option_strings)))

        if local_vars:
            r  = 'local %s' % ' '.join(f'{s}=()' for s in local_vars)
            for cmd in set_cmds:
                r += '\n%s' % cmd

            return r

        return ''

    def _generate_subcommand_call(self):
        if not self.subcommands:
            return ''

        query = self.query.use('get_positional')
        positional_num = self.subcommands.get_positional_num()

        r = 'case "$(%s get_positional %d)" in\n' % (query, positional_num)
        for subcommand in self.subcommands.subcommands:
            sub_funcname = self.ctxt.helpers.make_completion_funcname(subcommand)
            cmds = utils.get_all_command_variations(subcommand)
            pattern = '|'.join(shell.quote(s) for s in cmds)
            r += f'  {pattern}) {sub_funcname}; return $?;;\n'
        r += 'esac'

        return r

    def _generate_option_parsing(self):
        args = []

        if self.commandline.inherit_options:
            options = self.commandline.get_options(with_parent_options=True)
        else:
            options = self.commandline.get_options()

        for option in options:
            args.append(self._complete_option(option))

        for cmdline in self.commandline.get_parents():
            for option in cmdline.get_positionals():
                args.append(self._complete_positional(option))

            if cmdline.get_subcommands():
                args.append(self._complete_subcommands(cmdline.get_subcommands()))

        for option in self.commandline.get_positionals():
            args.append(self._complete_positional(option))

        if self.subcommands:
            args.append(self._complete_subcommands(self.subcommands))

        if self.commandline.wraps:
            args.append(Arg(None, None, "'::*'"))

        if not args:
            return ''

        args_with_when = []
        args_without_when = []
        for arg in args:
            if arg.when is None:
                args_without_when.append(arg)
            else:
                args_with_when.append(arg)

        r = ''

        if not args_without_when:
            r += 'local -a args=()\n'
        else:
            r += 'local -a args=(\n'
            for arg in args_without_when:
                r += '  %s\n' % arg.option_spec
            r += ')\n'

        for arg in args_with_when:
            when_cmd = zsh_when.generate_when_conditions(self.query, arg.when)
            r += '%s &&\\\n' % when_cmd
            r += '  args+=(%s)\n' % arg.option_spec

        if self.ctxt.config.option_stacking:
            r += '_arguments -S -s -w "${args[@]}"'
        else:
            r += '_arguments -S "${args[@]}"'

        return r

    def get_code(self):
        '''Return the code of the completion function.'''

        return '%s() {\n%s\n}' % (
            self.funcname,
            indent('\n\n'.join(c for c in self.code.values() if c), 2))


def get_zstyles(commandline, out_list):
    '''Get zstyle commands for disabling sorting completion suggestions.'''

    prog = commandline.get_program_name()

    for option in filter(lambda o: o.nosort, commandline.options):
        for option_string in option.option_strings:
            r = f"zstyle ':completion:*:*:{prog}:option{option_string}-*' sort false"
            out_list.append(r)

    for positional in filter(lambda p: p.nosort, commandline.positionals):
        number = positional.number
        r = f"zstyle ':completion:*:*:{prog}:argument-{number}:*' sort false"
        out_list.append(r)


def generate_completion(commandline, config=None):
    '''Code for generating a Zsh auto completion file.'''

    if config is None:
        config = config_.Config()

    commandline = generation.enhance_commandline(commandline, config)
    helpers = zsh_helpers.ZshHelpers(config, commandline.prog)
    ctxt = generation.GenerationContext(config, helpers)
    functions = generation.visit_commandlines(ZshCompletionFunction, ctxt, commandline)
    all_progs = ' '.join([commandline.prog] + commandline.aliases)

    if helpers.is_used('query'):
        types = utils.get_defined_option_types(commandline)
        if types.short:
            ctxt.helpers.use_function('query', 'short_options')
        if types.long:
            ctxt.helpers.use_function('query', 'long_options')
        if types.old:
            ctxt.helpers.use_function('query', 'old_options')

    output = Output(config, helpers)

    if config.zsh_compdef:
        output.add('#compdef %s' % all_progs)

    output.add_generation_notice()
    output.add_comments()
    output.add_included_files()

    completion_func, wrapper_code = zsh_wrapper.generate_wrapper(ctxt, commandline)

    output.add_helper_functions_code()

    output.extend(function.get_code() for function in functions)

    zstyles = []
    commandline.visit_commandlines(lambda c: get_zstyles(c, zstyles))
    if zstyles:
        output.add('\n'.join(zstyles))

    if wrapper_code:
        output.add(wrapper_code)

    if config.zsh_compdef:
        output.add('%s "$@"' % completion_func)
    else:
        output.add('compdef %s %s' % (completion_func, all_progs))

    output.add_vim_modeline('zsh')

    return output.get()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Functions for parsing the `when` attribute.'''

from . import shell_parser
from .errors import CrazyError


# pylint: disable=too-few-public-methods


class OptionIs:
    '''Class for holding `option_is`.'''

    def __init__(self, args):
        self.options = []
        self.values = []
        has_end_of_options = False

        for arg in args:
            if arg == '--':
                has_end_of_options = True
            elif not has_end_of_options:
                self.options.append(arg)
            elif has_end_of_options:
                self.values.append(arg)

        if not self.options:
            raise CrazyError('OptionIs: Empty options')

        if not self.values:
            raise CrazyError('OptionIs: Empty values')


class HasOption:
    '''Class for holding `has_option`.'''

    def __init__(self, args):
        self.options = args

        if not self.options:
            raise CrazyError('HasOption: Empty options')


def replace_commands(tokens):
    '''Replaced `shell_parser.Command` objects by condition objects.'''

    r = []

    for token in tokens:
        if isinstance(token, shell_parser.Command):
            cmd, *args = token.args

            if cmd == 'option_is':
                r.append(OptionIs(args))
            elif cmd == 'has_option':
                r.append(HasOption(args))
            else:
                raise CrazyError(f"Invalid command: {cmd!r}")
        else:
            r.append(token)

    return r


def needs_braces(tokens):
    '''Check if we need braces around our tokens.

    We don't need braces if:
        - We only have one token (single command)
        - We only have two tokens (negated single command)
        - We don't have an OR inside the tokens
    '''
    return len(tokens) > 2 and '||' in tokens


def parse_when(s):
    '''Parse `when` string and return an object.'''

    tokens = shell_parser.parse(s)
    return replace_commands(tokens)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Shell utility functions.'''

import re

from . import utils


def make_identifier(string):
    '''Make `string` a valid shell identifier.

    This function replaces any dashes '-' with underscores '_',
    removes any characters that are not letters, digits, or underscores,
    and ensures that consecutive underscores are replaced with a single
    underscore.

    Args:
        string (str):
            The input string to be converted into a valid shell identifier.

    Returns:
        str: The modified string that is a valid shell identifier.
    '''

    string = string.replace('-', '_')
    string = re.sub('[^a-zA-Z0-9_]', '', string)
    string = re.sub('_+', '_', string)
    if string and string[0] in '0123456789':
        return '_' + string
    return string


def needs_quote(string):
    '''Return if string needs quoting.'''

    return not re.fullmatch('[a-zA-Z0-9_@%+=:,./-]+', string)


def quote(string, quote_empty_string=True):
    '''Quotes a string for safe usage in shell commands or scripts.

    Args:
        string (str):
            The input string to be quoted.

        quote_empty_string (bool, optional):
            Determines whether to quote an empty string or not.
            Defaults to True.

    Returns:
        str: The quoted string.
    '''

    if not string and not quote_empty_string:
        return ''

    if not needs_quote(string):
        return string

    if "'" not in string:
        return f"'{string}'"

    if '"' not in string:
        s = string
        s = s.replace('\\', '\\\\')
        s = s.replace('$', '\\$')
        s = s.replace('`', '\\`')
        return f'"{s}"'

    s = string.replace("'", '\'"\'"\'')
    return f"'{s}'"


def join_quoted(arguments, string=' '):
    '''Joins quoted `arguments` on `string`.'''

    return string.join(quote(arg) for arg in arguments)


class ShellCompleter:
    '''Base class for argument completion.'''

    # pylint: disable=missing-function-docstring
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def complete(self, ctxt, trace, command, *a):
        if not hasattr(self, command):
            utils.warn(f"complete: Falling back from `{command}` to `none`")
            command = 'none'

        return getattr(self, command)(ctxt, trace, *a)

    def complete_from_def(self, ctxt, trace, definition):
        command, *args = definition
        return self.complete(ctxt, trace, command, *args)

    def fallback(self, ctxt, trace, from_, to, *a):
        utils.warn(f"ShellCompleter: Falling back from `{from_}` to `{to}`")
        return self.complete(ctxt, trace, to, *a)

    def signal(self, ctxt, trace):
        signals = {
            'ABRT':   'Process abort signal',
            'ALRM':   'Alarm clock',
            'BUS':    'Access to an undefined portion of a memory object',
            'CHLD':   'Child process terminated: stopped: or continued',
            'CONT':   'Continue executing: if stopped',
            'FPE':    'Erroneous arithmetic operation',
            'HUP':    'Hangup',
            'ILL':    'Illegal instruction',
            'INT':    'Terminal interrupt signal',
            'KILL':   'Kill (cannot be caught or ignored)',
            'PIPE':   'Write on a pipe with no one to read it',
            'QUIT':   'Terminal quit signal',
            'SEGV':   'Invalid memory reference',
            'STOP':   'Stop executing (cannot be caught or ignored)',
            'TERM':   'Termination signal',
            'TSTP':   'Terminal stop signal',
            'TTIN':   'Background process attempting read',
            'TTOU':   'Background process attempting write',
            'USR1':   'User-defined signal 1',
            'USR2':   'User-defined signal 2',
            'POLL':   'Pollable event',
            'PROF':   'Profiling timer expired',
            'SYS':    'Bad system call',
            'TRAP':   'Trace/breakpoint trap',
            'XFSZ':   'File size limit exceeded',
            'VTALRM': 'Virtual timer expired',
            'XCPU':   'CPU time limit exceeded',
        }

        return self.complete(ctxt, trace, 'choices', signals)

    def exec(self, _ctxt, _trace, _command):
        raise NotImplementedError

    def list(self, _ctxt, _trace, _command, _opts=None):
        raise NotImplementedError

    # =========================================================================
    # Aliases
    # =========================================================================

    def file_list(self, ctxt, trace, opts=None):
        list_opts = {
            'separator': opts.pop('separator', ',') if opts else ',',
            'duplicates': opts.pop('duplicates', False) if opts else False
        }

        return self.list(ctxt, trace, ['file', opts], list_opts)

    def directory_list(self, ctxt, trace, opts=None):
        list_opts = {
            'separator': opts.pop('separator', ',') if opts else ',',
            'duplicates': opts.pop('duplicates', False) if opts else False
        }

        return self.list(ctxt, trace, ['directory', opts], list_opts)

    # =========================================================================
    # Bonus
    # =========================================================================

    def login_shell(self, ctxt, trace):
        return self.exec(ctxt, trace, "command grep -E '^[^#]' /etc/shells")

    def locale(self, ctxt, trace):
        return self.exec(ctxt, trace, "command locale -a")

    def charset(self, ctxt, trace):
        return self.exec(ctxt, trace, "command locale -m")



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains Exception classes for crazy-complete.'''


class CrazyError(Exception):
    '''Exception class for handling predictable or expected errors.

    This exception is raised in situations where the error condition is
    anticipated and should be handled gracefully by the program. It is
    meant to signal issues that are not caused by bugs in the code but
    by user input.
    '''


class CrazyTypeError(CrazyError):
    '''Exception raised for invalid parameter types.

    Args:
        name (str): The name of the parameter with the invalid type.
        expected (str): A description of the expected types for the parameter.
        value (any): The actual value passed that has the wrong type.
    '''

    def __init__(self, name, expected, value):
        self.name = name
        self.expected = expected
        self.value = value
        super().__init__(self.__str__())

    def __str__(self):
        s0 = 'Parameter `%s` has an invalid type.' % self.name
        s1 = 'Expected types: %s. Received: %r (%s)' % (
            self.expected, self.value, type(self.value).__name__)
        return f'{s0} {s1}'


class CrazySchemaValidationError(CrazyError):
    '''Exception raised for errors in the configuration structure.

    This exception is specifically designed to handle errors encountered when
    validating YAML or JSON configuration files against a defined schema.
    It provides detailed error messages, including the line and column numbers
    where the error occurred, to help locate the issue in the configuration
    file.

    Args:
        message (str): The error message.
        value_with_trace (ValueWithTrace): The value that caused the error.
    '''

    def __init__(self, message, value_with_trace):
        self.message = message
        self.value_with_trace = value_with_trace
        super().__init__(self.__str__())

    def __str__(self):
        pos_string = self.value_with_trace.get_position_string()

        if not pos_string:
            return self.message

        return f'{pos_string}: {self.message}'


class InternalError(Exception):
    '''Exception raised for internal errors within the program.

    This exception indicates that an unexpected condition has occurred that
    typically results from a bug or flaw in the code logic. It represents
    errors that are not caused by user actions or external factors, but
    rather by mistakes or inconsistencies in the program's internal state.
    '''



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains code for completing arguments in Bash.'''

from . import shell
from . import bash_versions
from .str_utils import indent
from .type_utils import is_dict_type
from .bash_utils import make_file_extension_pattern


class BashCompletionBase:
    '''Base class for Bash completions.'''

    def get_code(self, append=False):
        '''Returns the code used for completing an argument.

        If `append` is `True`, then the results will be appended to $COMPREPLY,
        otherwise $COMPREPLY will be truncated.

        Args:
            append (bool): If True, the results will be appended to $COMPREPLY.

        Returns:
            str: The command for Bash completion.
        '''
        raise NotImplementedError

    def get_function(self):
        '''Returns a function.'''
        raise NotImplementedError


class BashCompletionCode(BashCompletionBase):
    '''Used for completion code that internally modify $COMPREPLY.'''

    def __init__(self, ctxt, code):
        self.ctxt = ctxt
        self.code = code

    def get_code(self, append=False):
        if not self.code:
            return ''

        r = []

        if append:
            r.append('local COMPREPLY_OLD=("${COMPREPLY[@]}")')

        r.append(self.code)

        if append:
            r.append('COMPREPLY=("${COMPREPLY_OLD[@]}" "${COMPREPLY[@]}")')

        return '\n'.join(r)

    def get_function(self):
        if not self.code:
            return 'true'

        return self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())


class BashCompletionFunc(BashCompletionBase):
    '''Used for completion functions that internally modify $COMPREPLY.'''

    def __init__(self, ctxt, args):
        self.ctxt = ctxt
        self.args = args

    def get_code(self, append=False):
        r = []

        if append:
            r.append('local COMPREPLY_OLD=("${COMPREPLY[@]}")')

        r.append(shell.join_quoted(self.args))

        if append:
            r.append('COMPREPLY=("${COMPREPLY_OLD[@]}" "${COMPREPLY[@]}")')

        return '\n'.join(r)

    def get_function(self):
        if len(self.args) == 1:
            return self.args[0]

        return self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())


class CompgenW(BashCompletionBase):
    '''Used for completing a list of words.'''

    def __init__(self, ctxt, values):
        self.ctxt = ctxt
        self.values = [str(value) for value in values]

    def get_code(self, append=False):
        needs_quote = any(filter(shell.needs_quote, self.values))

        if not needs_quote:
            return 'COMPREPLY%s=($(compgen -W %s -- "$cur"))' % (
                ('+' if append else ''),
                shell.quote(' '.join(self.values)))

        return ('%s %s-- %s' % (
            self.ctxt.helpers.use_function('values'),
            ('-a ' if append else ''),
            shell.join_quoted(self.values)))

    def get_function(self):
        return self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())


class BashCompletionCompgen(BashCompletionBase):
    '''Used for completion that uses Bash's `compgen` command.'''

    def __init__(self, ctxt, compgen_args, code=None):
        self.ctxt = ctxt
        self.compgen_args = compgen_args
        self.code = code

    def get_code(self, append=False):
        compgen_cmd = 'COMPREPLY%s=($(compgen %s -- "$cur"))' % (
            ('+' if append else ''), self.compgen_args)

        if not self.code:
            return compgen_cmd

        return f'{self.code}\n{compgen_cmd}'

    def get_function(self):
        return self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())


class BashCompleteCombine(BashCompletionBase):
    '''Used for combining multiple completion commands.'''

    def __init__(self, ctxt, trace, completer, commands):
        self.ctxt = ctxt
        self.completion_objects = []

        for command_args in commands:
            obj = completer.complete_from_def(ctxt, trace, command_args)
            self.completion_objects.append(obj)

    def get_code(self, append=False):
        code = [self.completion_objects[0].get_code(append=append)]
        for obj in self.completion_objects[1:]:
            code.append(obj.get_code(append=True))
        return '\n'.join(code)

    def get_function(self):
        return self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())


class BashCompleteKeyValueList(BashCompletionCode):
    '''Used for completing a list of key-value pairs.'''

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def __init__(self, ctxt, trace, completer,
                 pair_separator, value_separator, values):
        funcs = {}

        for key, _desc, complete in values:
            if not complete:
                funcs[key] = 'false'
            elif complete[0] == 'none':
                funcs[key] = 'true'
            else:
                obj = completer.complete_from_def(ctxt, trace, complete)
                funcs[key] = obj.get_function()

        q = shell.quote
        args = ' \\\n'.join('%s %s' % (q(k), q(f)) for k, f in funcs.items())

        code = '%s %s %s \\\n%s' % (
            ctxt.helpers.use_function('key_value_list'),
            shell.quote(pair_separator),
            shell.quote(value_separator),
            indent(args, 2)
        )

        super().__init__(ctxt, code)


class BashCompleter(shell.ShellCompleter):
    '''Code generator for completing arguments in Bash.'''

    # pylint: disable=missing-function-docstring
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def none(self, ctxt, _trace, *_):
        return BashCompletionCode(ctxt, '')

    def integer(self, ctxt, _trace, options=None):
        suffixes = []

        if options:
            if 'suffixes' in options:
                for suffix, description in options['suffixes'].items():
                    suffixes.append(f'{suffix}:{description}')

        if not suffixes:
            return BashCompletionCode(ctxt, '')

        func = ctxt.helpers.use_function('number')
        return BashCompletionFunc(ctxt, [func, *suffixes])

    def float(self, ctxt, trace, options=None):
        return self.integer(ctxt, trace, options)

    def choices(self, ctxt, _trace, choices):
        return CompgenW(ctxt, choices)

    def command(self, ctxt, _trace, opts=None):
        code = None
        path = None
        append = None
        prepend = None

        if opts:
            path = opts.get('path', None)
            append = opts.get('path_append', None)
            prepend = opts.get('path_prepend', None)

        if path:
            code = 'local -x PATH=%s' % shell.quote(path)
        elif append and prepend:
            append = shell.quote(append)
            prepend = shell.quote(prepend)
            code = 'local -x PATH=%s:"$PATH":%s' % (prepend, append)
        elif append:
            code = 'local -x PATH="$PATH":%s' % shell.quote(append)
        elif prepend:
            code = 'local -x PATH=%s:"$PATH"' % shell.quote(prepend)

        return BashCompletionCompgen(ctxt, '-A command', code=code)

    def directory(self, ctxt, _trace, opts=None):
        directory = None if opts is None else opts.get('directory', None)
        filedir = bash_versions.filedir(ctxt)

        if directory:
            r = 'builtin pushd %s &>/dev/null && {\n' % shell.quote(directory)
            r += '  %s -d\n' % filedir
            r += '  builtin popd >/dev/null\n'
            r += '}'
            return BashCompletionCode(ctxt, r)

        return BashCompletionFunc(ctxt, [filedir, '-d'])

    def file(self, ctxt, _trace, opts=None):
        fuzzy = False
        directory = None
        extensions = None
        ignore_globs = None

        if opts:
            fuzzy = opts.get('fuzzy', False)
            directory = opts.get('directory', None)
            extensions = opts.get('extensions', None)
            ignore_globs = opts.get('ignore_globs', None)

        args = [bash_versions.filedir(ctxt)]

        if extensions:
            args.append(make_file_extension_pattern(extensions, fuzzy))

        if not directory and not ignore_globs:
            return BashCompletionFunc(ctxt, args)

        if directory:
            r = 'builtin pushd %s &>/dev/null && {\n' % shell.quote(directory)
            r += '  %s\n' % shell.join_quoted(args)
            r += '  builtin popd >/dev/null\n'
            r += '}'
        else:
            r = shell.join_quoted(args)

        if ignore_globs:
            func = ctxt.helpers.use_function('file_filter')
            r += '\n%s %s' % (func, shell.join_quoted(ignore_globs))

        return BashCompletionCode(ctxt, r)

    def mime_file(self, ctxt, _trace, pattern):
        func = ctxt.helpers.use_function('mime_file')
        return BashCompletionFunc(ctxt, [func, pattern])

    def group(self, ctxt, _trace):
        return BashCompletionCompgen(ctxt, '-A group')

    def hostname(self, ctxt, _trace):
        return BashCompletionCompgen(ctxt, '-A hostname')

    def pid(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.pids(ctxt)])

    def process(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.pnames(ctxt)])

    def range(self, ctxt, _trace, start, stop, step=1):
        if step == 1:
            return BashCompletionCompgen(ctxt, f"-W '{{{start}..{stop}}}'")

        return BashCompletionCompgen(ctxt, f"-W '{{{start}..{stop}..{step}}}'")

    def service(self, ctxt, _trace):
        return BashCompletionCompgen(ctxt, '-A service')

    def user(self, ctxt, _trace):
        return BashCompletionCompgen(ctxt, '-A user')

    def variable(self, ctxt, _trace):
        return BashCompletionCompgen(ctxt, '-A variable')

    def environment(self, ctxt, _trace):
        return BashCompletionCompgen(ctxt, '-A export')

    def exec(self, ctxt, _trace, command):
        func = ctxt.helpers.use_function('exec')
        return BashCompletionFunc(ctxt, [func, command])

    def exec_fast(self, ctxt, _trace, command):
        func = ctxt.helpers.use_function('exec_fast')
        return BashCompletionFunc(ctxt, [func, command])

    def exec_internal(self, ctxt, _trace, command):
        return BashCompletionCode(ctxt, command)

    def value_list(self, ctxt, _trace, opts):
        separator = opts.get('separator', ',')
        duplicates = opts.get('duplicates', False)
        values = opts['values']

        if is_dict_type(values):
            values = list(values.keys())

        args = [ctxt.helpers.use_function('value_list')]
        if duplicates:
            ctxt.helpers.use_function('value_list', 'duplicates')
            args.append('-d')
        args.append(separator)

        return BashCompletionFunc(ctxt, args + values)

    def key_value_list(self, ctxt, trace, pair_separator, value_separator, values):
        return BashCompleteKeyValueList(ctxt, trace, self, pair_separator, value_separator, values)

    def combine(self, ctxt, trace, commands):
        return BashCompleteCombine(ctxt, trace, self, commands)

    def list(self, ctxt, trace, command, opts=None):
        separator = opts.get('separator', ',') if opts else ','
        duplicates = opts.get('duplicates', False) if opts else False

        obj = self.complete_from_def(ctxt, trace, command)
        code = obj.get_code()

        func = ctxt.helpers.use_function('value_list')

        dup_arg = ''
        if duplicates:
            ctxt.helpers.use_function('value_list', 'duplicates')
            dup_arg = ' -d'

        r = 'local cur_old="$cur"\n'
        r += 'cur="${cur##*%s}"\n' % shell.quote(separator)
        r += '%s\n' % code
        r += 'cur="$cur_old"\n'
        r += '%s%s %s "${COMPREPLY[@]}"' % (func, dup_arg, shell.quote(separator))
        return BashCompletionCode(ctxt, r)

    def history(self, ctxt, _trace, pattern):
        func = ctxt.helpers.use_function('history')
        return BashCompletionFunc(ctxt, [func, pattern])

    def commandline_string(self, ctxt, _trace):
        func = ctxt.helpers.use_function('commandline_string')
        return BashCompletionFunc(ctxt, [func])

    def command_arg(self, ctxt, _trace):
        return BashCompletionCode(ctxt, '')

    def date(self, ctxt, _trace, _format):
        return BashCompletionCode(ctxt, '')

    def date_format(self, ctxt, _trace):
        return BashCompletionCode(ctxt, '')

    def uid(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.uids(ctxt)])

    def gid(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.gids(ctxt)])

    def signal(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.signals(ctxt)])

    def filesystem_type(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.fstypes(ctxt)])

    def prefix(self, ctxt, trace, prefix, command):
        obj = self.complete_from_def(ctxt, trace, command)
        func = obj.get_function()
        prefix_func = ctxt.helpers.use_function('prefix')
        return BashCompletionFunc(ctxt, [prefix_func, prefix, func])

    def ip_address(self, ctxt, _trace, type_='all'):
        func = bash_versions.ip_addresses(ctxt)

        if type_ == 'ipv6':
            return BashCompletionFunc(ctxt, [func, '-6'])

        if type_ == 'ipv4':
            return BashCompletionFunc(ctxt, [func])

        return BashCompletionFunc(ctxt, [func, '-a'])

    # =========================================================================
    # Bonus
    # =========================================================================

    def mountpoint(self, ctxt, _trace):
        func = ctxt.helpers.use_function('mountpoint')
        return BashCompletionFunc(ctxt, [func])

    def locale(self, ctxt, trace):
        func = ctxt.helpers.use_function('locales')
        return BashCompletionFunc(ctxt, [func])

    def charset(self, ctxt, trace):
        func = ctxt.helpers.use_function('charsets')
        return BashCompletionFunc(ctxt, [func])

    def login_shell(self, ctxt, _trace):
        return BashCompletionFunc(ctxt, [bash_versions.shells(ctxt)])

    def net_interface(self, ctxt, _trace):
        func = bash_versions.available_interfaces(ctxt)
        return BashCompletionFunc(ctxt, [func])

    def timezone(self, ctxt, _trace):
        exec_func = ctxt.helpers.use_function('exec')
        list_func = ctxt.helpers.use_function('timezone_list')
        return BashCompletionFunc(ctxt, [exec_func, list_func])

    def alsa_card(self, ctxt, _trace):
        func = ctxt.helpers.use_function('alsa_list_cards')
        return BashCompletionFunc(ctxt, [func])

    def alsa_device(self, ctxt, _trace):
        func = ctxt.helpers.use_function('alsa_list_devices')
        return BashCompletionFunc(ctxt, [func])



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Algorithm functions.'''

from collections import OrderedDict


def flatten(iterable):
    '''Flatten a list of iterables into a single list.'''

    result = []
    for item in iterable:
        result.extend(item)
    return result


def uniq(iterable):
    '''Return list of unique items, preserving order.'''

    result = []
    seen = set()
    for item in iterable:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result


def partition(iterable, predicate):
    '''Split an iterable into two lists based on a predicate.

    Args:
        iterable:
            The input iterable to split.
        predicate (callable):
            A function returning True or False for each element.

    Returns:
        tuple: (matched, unmatched) where
            matched   -- list of items for which predicate(item) is True
            unmatched -- list of items for which predicate(item) is False
    '''

    a, b = [], []
    for item in iterable:
        if predicate(item):
            a.append(item)
        else:
            b.append(item)
    return (a, b)


def group_by(iterable, keyfunc):
    '''Group items from an iterable by a key function.

    Args:
        iterable:
            The input iterable of items to group.
        keyfunc (callable):
            Function applied to each item to determine its group key.

    Returns:
        OrderedDict:
            A mapping of key -> list of items with that key, preserving
            the order of first appearance of each key.
    '''
    result = OrderedDict()

    for item in iterable:
        val = keyfunc(item)

        if val in result:
            result[val].append(item)
        else:
            result[val] = [item]

    return result


def numbers_are_contiguous(numbers):
    '''Return True if the list of integers is contiguous (1, 2, 3, 4, ...)'''

    if not numbers:
        return False

    sorted_nums = sorted(numbers)
    return all(b - a == 1 for a, b in zip(sorted_nums, sorted_nums[1:]))



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains helper functions for Zsh.'''

from .helpers import GeneralHelpers, ShellFunction


_QUERY = ShellFunction('query', r'''
# ===========================================================================
#
# This function is for querying the command line.
#
# COMMANDS
#   init <OPTIONS> <ARGS...>
#     This is the first call you have to make, otherwise the other commands
#     won't (successfully) work.
#
#     It parses <ARGS> according to <OPTIONS> and stores results in the
#     variables POSITIONALS, HAVING_OPTIONS and OPTION_VALUES.
#
#     The first argument is a comma-separated list of options that the parser
#     should know about. Short options (-o), long options (--option), and
#     old-style options (-option) are supported.
#
#     If an option takes an argument, it is suffixed by '='.
#     If an option takes an optional argument, it is suffixed by '=?'.
#
#   get_positional <NUM>
#     Prints out the positional argument number NUM (starting from 1)
#
#   has_option [WITH_INCOMPLETE] <OPTIONS...>
#     Checks if an option given in OPTIONS is passed on commandline.
#     If an option requires an argument, this command returns true only if the
#     option includes an argument. If 'WITH_INCOMPLETE' is specified, it also
#     returns true for options missing their arguments.
#
#   option_is <OPTIONS...> -- <VALUES...>
#     Checks if one option in OPTIONS has a value of VALUES.
#
# EXAMPLE
#   local POSITIONALS HAVING_OPTIONS OPTION_VALUES
#   query init '-f,-a=,-optional=?' program_name -f -optional -a foo bar
#   query has_option -f
#   query option_is -a -- foo
#
#   Here, -f is a flag, -a takes an argument, and -optional takes an optional
#   argument.
#
#   Both queries return true.
#
# ===========================================================================

__zsh_query_contains() {
  local arg='' key="$1"; shift
  for arg; do [[ "$key" == "$arg" ]] && return 0; done
  return 1
}

local cmd="$1"; shift

case "$cmd" in
#ifdef get_positional
  get_positional)
#ifdef DEBUG
    if (( $# != 1 )); then
      echo "%FUNCNAME%: get_positional: takes exactly one argument" >&2
      return 1
    fi

    if test "$1" -eq 0; then
      echo "%FUNCNAME%: get_positional: positionals start at 1, not 0!" >&2
      return 1
    fi

#endif
    printf "%s" "${POSITIONALS[$1]}"
    return 0;;
#endif
#ifdef get_option
  get_option)
    local i=0

#ifdef DEBUG
    if (( $# == 0 )); then
      echo "%FUNCNAME%: get_option: arguments required" >&2
      return 1
    fi

#endif
    for (( i=0; i <= ${#HAVING_OPTIONS[@]}; ++i )); do
      local option="${HAVING_OPTIONS[$i]}"
      if __zsh_query_contains "$option" "$@"; then
        printf "%s\n" "${OPTION_VALUES[$i]}"
      fi
    done

    return 0;;
#endif
#ifdef has_option
  has_option)
#ifdef with_incomplete
    local option='' with_incomplete=0
    [[ "$1" == 'WITH_INCOMPLETE' ]] && { with_incomplete=1; shift; }
#else
    local option=''
#endif

#ifdef DEBUG
    if (( $# == 0 )); then
      echo "%FUNCNAME%: has_option: arguments required" >&2
      return 1
    fi

#endif
    for option in "${HAVING_OPTIONS[@]}"; do
      __zsh_query_contains "$option" "$@" && return 0
    done

#ifdef with_incomplete
    (( with_incomplete )) && \
      __zsh_query_contains "$INCOMPLETE_OPTION" "$@" && return 0

#endif
    return 1;;
#endif
#ifdef option_is
  option_is)
    local i=0 dash_dash_pos=0 cmd_option_is_options=() cmd_option_is_values=()

    dash_dash_pos=${@[(i)--]}
    cmd_option_is_options=("${@:1:$((dash_dash_pos - 1))}")
    cmd_option_is_values=("${@:$((dash_dash_pos + 1))}")

#ifdef DEBUG
    if (( ${#cmd_option_is_options[@]} == 0 )); then
      echo "%FUNCNAME%: option_is: missing options" >&2
      return 1
    fi

    if (( ${#cmd_option_is_values[@]} == 0 )); then
      echo "%FUNCNAME%: option_is: missing values" >&2
      return 1
    fi

#endif
    for (( i=${#HAVING_OPTIONS[@]}; i > 0; --i )); do
      local option="${HAVING_OPTIONS[$i]}"
      if __zsh_query_contains "$option" "${cmd_option_is_options[@]}"; then
        local value="${OPTION_VALUES[$i]}"
        __zsh_query_contains "$value" "${cmd_option_is_values[@]}" && return 0
      fi
    done

    return 1;;
#endif
  init)
    local -a options=(${=1})
    shift;;
  *)
    echo "%FUNCNAME%: argv[1]: invalid command" >&2
    return 1;;
esac

# continuing init...

# ===========================================================================
# Parsing of available options
# ===========================================================================

#ifdef old_options
local   old_opts_with_arg=()   old_opts_with_optional_arg=()   old_opts_without_arg=()
#endif
#ifdef long_options
local  long_opts_with_arg=()  long_opts_with_optional_arg=()  long_opts_without_arg=()
#endif
#ifdef short_options
local short_opts_with_arg=() short_opts_with_optional_arg=() short_opts_without_arg=()
#endif

local option=''
for option in "${options[@]}"; do
  case "$option" in
#ifdef long_options
    --?*=)    long_opts_with_arg+=("${option%=}");;
    --?*=\?)  long_opts_with_optional_arg+=("${option%=?}");;
    --?*)     long_opts_without_arg+=("$option");;
#endif
#ifdef short_options
    -?=)      short_opts_with_arg+=("${option%=}");;
    -?=\?)    short_opts_with_optional_arg+=("${option%=?}");;
    -?)       short_opts_without_arg+=("$option");;
#endif
#ifdef old_options
    -??*=)    old_opts_with_arg+=("${option%=}");;
    -??*=\?)  old_opts_with_optional_arg+=("${option%=?}");;
    -??*)     old_opts_without_arg+=("$option");;
#endif
#ifdef DEBUG
    *) echo "%FUNCNAME%: $option: not a valid short, long or oldstyle option" >&2; return 1;;
#endif
  esac
done

# ===========================================================================
# Parsing of command line options
# ===========================================================================

POSITIONALS=()
HAVING_OPTIONS=()
OPTION_VALUES=()
INCOMPLETE_OPTION=''

local args=("${(Q)@}")
local argi=2 # argi[1] is program name
for ((; argi <= ${#args[@]}; ++argi)); do
  local arg="${args[$argi]}"
  local have_trailing_arg=$(test $argi -lt $# && echo true || echo false)

  case "$arg" in
    --)
      POSITIONALS+=("${@:$((argi + 1))}")
      break;;
    --*=*)
      HAVING_OPTIONS+=("${arg%%=*}")
      OPTION_VALUES+=("${arg#*=}");;
    --*)
#ifdef long_options
      if __zsh_query_contains "$arg" "${long_opts_with_arg[@]}"; then
        if $have_trailing_arg; then
          HAVING_OPTIONS+=("$arg")
          OPTION_VALUES+=("${args[$((++argi))]}")
#ifdef with_incomplete
        else
          INCOMPLETE_OPTION="$arg"
#endif
        fi
      else
        HAVING_OPTIONS+=("$arg")
        OPTION_VALUES+=("")
      fi
#endif
      ;;
    -?*) # ignore '-'
#ifdef old_options
      if [[ "$arg" == *=* ]]; then
        local option="${arg%%=*}" value="${arg#*=}"
        if __zsh_query_contains "$option" "${old_opts_with_arg[@]}" "${old_opts_with_optional_arg[@]}"; then
          HAVING_OPTIONS+=("$option")
          OPTION_VALUES+=("$value")
          continue
        fi
      elif __zsh_query_contains "$arg" "${old_opts_with_arg[@]}"; then
        if $have_trailing_arg; then
          HAVING_OPTIONS+=("$arg")
          OPTION_VALUES+=("${args[$((++argi))]}")
#ifdef with_incomplete
        else
          INCOMPLETE_OPTION="$arg"
#endif
        fi
        continue
      elif __zsh_query_contains "$arg" "${old_opts_without_arg[@]}" "${old_opts_with_optional_arg[@]}"; then
        HAVING_OPTIONS+=("$arg")
        OPTION_VALUES+=("")
        continue
      fi
#endif

#ifdef short_options
      local i=1 arg_length=${#arg}
      for ((; i < arg_length; ++i)); do
        local option="-${arg:$i:1}"
        local trailing_chars="${arg:$((i+1))}"

        if __zsh_query_contains "$option" "${short_opts_without_arg[@]}"; then
          HAVING_OPTIONS+=("$option")
          OPTION_VALUES+=("")
        elif __zsh_query_contains "$option" "${short_opts_with_arg[@]}"; then
          if [[ -n "$trailing_chars" ]]; then
            HAVING_OPTIONS+=("$option")
            OPTION_VALUES+=("$trailing_chars")
          elif $have_trailing_arg; then
            HAVING_OPTIONS+=("$option")
            OPTION_VALUES+=("${args[$((++argi))]}")
#ifdef with_incomplete
          else
            INCOMPLETE_OPTION="$option"
#endif
          fi

          continue 2
        elif __zsh_query_contains "$option" "${short_opts_with_optional_arg[@]}"; then
          HAVING_OPTIONS+=("$option")
          OPTION_VALUES+=("$trailing_chars") # may be empty
          continue 2
        fi
      done
#endif
      ;;
    *)
      POSITIONALS+=("$arg");;
  esac
done
''')

_PREFIX = ShellFunction('prefix', r'''
if [[ "$PREFIX" == "$1"* ]]; then
  PREFIX="${PREFIX#"$1"}"
  IPREFIX="$IPREFIX$1"
  $2
else
  compadd -- "$1"
fi
''')

_EXEC = ShellFunction('exec', r'''
local item='' desc='' describe=()

while IFS=$'\t' read -r item desc; do
  item="${item//:/\\:}"
  [[ -n "$desc" ]] && describe+=("$item:$desc") || describe+=("$item")
done < <(eval "$1")

_describe '' describe
''')

_HISTORY = ShellFunction('history', r'''
[[ -f "$HISTFILE" ]] || return

local match=''

command grep -E -o -- "$1" "$HISTFILE" | while read -r match; do
  compadd -- "$match"
done
''')

_MIME_FILE = ShellFunction('mime_file', r'''
local line='' file='' mime='' i_opt=''

if command file -i /dev/null &>/dev/null; then
  i_opt="-i"
elif command file -I /dev/null &>/dev/null; then
  i_opt="-I"
else
  compadd -- "$PREFIX"*
  return
fi

command file -L $i_opt -- "$PREFIX"* 2>/dev/null | while read -r line; do
  mime="${line##*:}"

  if [[ "$mime" == *inode/directory* ]] || command grep -q -E -- "$1" <<< "$mime"; then
    file="${line%:*}"

    if [[ "$file" == *\\* ]]; then
      file="$(command perl -pe 's/\\([0-7]{3})/chr(oct($1))/ge' <<< "$file")"
    fi

    if [[ "$mime" == *inode/directory* ]]; then
      compadd -- "$file/"
    else
      compadd -- "$file"
    fi
  fi
done
''')

_UID_LIST = ShellFunction('uid_list', r'''
local items=($(command getent passwd | command awk -F: '{printf "%s:%s\n", $3, $1}'))
_describe 'users' items
''')

_GID_LIST = ShellFunction('gid_list', r'''
local items=($(command getent group | command awk -F: '{printf "%s:%s\n", $3, $1}'))
_describe 'groups' items
''')

_CHARSET_LIST = ShellFunction('charset_list', r'''
local items=($(command locale -m))
_describe 'charsets' items
''')

_PATH_FILES_RELATIVE = ShellFunction('path_files_relative', r'''
local DIR="$1"; shift
_path_files -W "$PWD/$DIR" "$@"
''')

# =============================================================================
# Bonus
# =============================================================================

_MOUNTPOINT = ShellFunction('mountpoint', r'''
local item=''

command findmnt -lno TARGET | while read -r item; do
  compadd -- "$item"
done
''')

_ALSA_COMPLETE_CARDS = ShellFunction('alsa_complete_cards', r'''
local card='' cards=()
command aplay -l \
  | command grep -Eo '^card [0-9]+: [^,]+' \
  | command uniq \
  | while builtin read card; do
  card="${card#card }"
  local id="${card%%: *}"
  local name="${card#*: }"
  cards+=("$id:$name")
done

_describe 'ALSA card' cards
''')

_ALSA_COMPLETE_DEVICES = ShellFunction('alsa_complete_devices', r'''
local card='' id='' name='' devices=()
command aplay -l \
  | command grep -Eo '^card [0-9]+: [^,]+' \
  | command uniq \
  | while builtin read card; do
  card="${card#card }"
  id="${card%%: *}"
  name="${card#*: }"
  devices+=("hw\\:$id:$name")
done

_describe 'ALSA device' devices
''')


class ZshHelpers(GeneralHelpers):
    '''Class holding helper functions for Zsh.'''

    def __init__(self, config, function_prefix):
        super().__init__(config, function_prefix, ShellFunction)
        self.add_function(_QUERY)
        self.add_function(_EXEC)
        self.add_function(_PREFIX)
        self.add_function(_HISTORY)
        self.add_function(_PATH_FILES_RELATIVE)
        self.add_function(_MIME_FILE)
        self.add_function(_UID_LIST)
        self.add_function(_GID_LIST)
        self.add_function(_CHARSET_LIST)
        self.add_function(_MOUNTPOINT)
        self.add_function(_ALSA_COMPLETE_CARDS)
        self.add_function(_ALSA_COMPLETE_DEVICES)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''
This module provides functions to retrieve file paths for
autocompletion scripts for Bash, Fish, and Zsh.
'''

import subprocess

from . import utils


# pylint: disable=broad-exception-caught
# pylint: disable=broad-exception-raised


def _pkg_config(args):
    '''
    Call the `pkg-config` program with `args` and check the return code.
    If everything is fine, return the output as string.
    '''

    command = ['pkg-config'] + args

    try:
        result = subprocess.run(
            command,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            text=True,
            check=False)
    except FileNotFoundError as exc:
        raise Exception('Program `pkg-config` not found') from exc

    if result.returncode == 0:
        return result.stdout.strip()

    raise Exception('Command `%s` failed: %s' % (
        ' '.join(command), result.stderr.strip()))


def get_bash_completion_file(program_name):
    '''Get the path for a Bash completion file.'''

    directory = '/usr/share/bash-completion/completions'

    try:
        directory = _pkg_config(['--variable=completionsdir', 'bash-completion'])
    except Exception as e:
        utils.warn(e)

    return f'{directory}/{program_name}'


def get_fish_completion_file(program_name):
    '''Get the path for a Fish completion file.'''

    directory = '/usr/share/fish/vendor_completions.d'

    try:
        directory = _pkg_config(['--variable=completionsdir', 'fish'])
    except Exception as e:
        utils.warn(e)

    return f'{directory}/{program_name}.fish'


def get_zsh_completion_file(program_name):
    '''Get the path for a Zsh completion file.'''

    # There is '/usr/share/zsh/vendor-completions', but that directory is not
    # in the default $fpath of zsh.
    directory = '/usr/share/zsh/site-functions'
    return f'{directory}/_{program_name}'



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains the configuration class.'''

import sys

from .type_utils import is_list_type


def _assert_is_bool(obj, func, param):
    if isinstance(obj, bool):
        return

    raise AssertionError(f"Config.{func}: {param}: expected bool, got `{obj}`")


class Config:
    '''Configuration settings for command line completion.'''

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods

    def __init__(self):
        self.function_prefix        = '_$PROG'
        self.debug                  = False
        self.abbreviate_commands    = False
        self.abbreviate_options     = False
        self.repeatable_options     = False
        self.inherit_options        = False
        self.option_stacking        = True
        self.long_opt_arg_sep       = 'both'
        self.vim_modeline           = True
        self.include_files          = []
        self.comments               = []
        self.keep_comments          = False
        self.line_length            = 80
        self.bash_completions_version = (2,)
        self.zsh_compdef            = True
        self.fish_fast              = False
        self.fish_inline_conditions = False

        self.disabled_hidden        = False
        self.disabled_final         = False
        self.disabled_groups        = False
        self.disabled_repeatable    = False
        self.disabled_when          = False

    def set_function_prefix(self, prefix):
        '''Sets the function prefix used for generated functions.

        Args:
            prefix (str):
                The prefix to be used. May contain the `$PROG` placeholder.

        Notes:
            This defaults to `_$PROG`
        '''

        assert isinstance(prefix, str)

        self.function_prefix = prefix

    def set_debug(self, enable):
        '''Sets debug mode.

        Args:
            enable (bool):
                If True, additional code for debugging is generated.

        Notes:
            This feature defaults to `False`.
        '''

        _assert_is_bool(enable, "set_debug", "enable")

        self.debug = enable

    def set_abbreviate_commands(self, enable):
        '''Sets whether commands can be abbreviated.

        Args:
            enable (bool):
                If True, commands can be abbreviated; if False, they cannot.

        Notes:
            This feature defaults to `False`.

            Implementation status for shells:
                Bash:
                    - set_abbreviate_commands(True): works
                    - set_abbreviate_commands(False): works
                Fish:
                    - set_abbreviate_commands(True): works
                    - set_abbreviate_commands(False): works
                Zsh:
                    - set_abbreviate_commands(True): works
                    - set_abbreviate_commands(False): works

        See Also:
            cli.CommandLine(..., abbreviate_commands=BOOL, ...)
        '''

        _assert_is_bool(enable, "set_abbreviate_commands", "enable")

        self.abbreviate_commands = enable

    def set_abbreviate_options(self, enable):
        '''Sets whether options can be abbreviated.

        Args:
            enable (bool):
                If True, options can be abbreviated; if False, they cannot.

        Notes:
            This feature defaults to `False`.

            Implementation status for shells:
                Bash:
                    - set_abbreviate_options(True): works
                    - set_abbreviate_options(False): works
                Fish:
                    - set_abbreviate_options(True): not implemented
                    - set_abbreviate_options(False): works
                Zsh:
                    - set_abbreviate_options(True): not implemented
                    - set_abbreviate_options(False): works

        See Also:
            cli.CommandLine(..., abbreviate_options=BOOL, ...)
        '''

        _assert_is_bool(enable, "set_abbreviate_options", "enable")

        self.abbreviate_options = enable

    def set_repeatable_options(self, enable):
        '''Sets whether options are suggested multiple times during completion.

        Args:
            enable (bool):
                If True, options can appear multiple times during completion;
                if False, options are suggested only once.

        Notes:
            This feature defaults to `False`.

            Implementation status for shells:
                Bash:
                    - set_repeatable_options(True): works
                    - set_repeatable_options(False): works
                Fish:
                    - set_repeatable_options(True): works
                    - set_repeatable_options(False): works
                Zsh:
                    - set_repeatable_options(True): works
                    - set_repeatable_options(False): works

        See Also:
            cli.CommandLine.add_option(..., repeatable=BOOL, ...)
        '''

        _assert_is_bool(enable, "set_repeatable_options", "enable")

        self.repeatable_options = enable

    def set_inherit_options(self, enable):
        '''Sets whether parent options are visible to subcommands.

        Args:
            enable (bool):
                If True, parent options are visible to subcommands.
                If False, they are not.

        Notes:
            This feature defaults to `False`.

            Implementation status for shells:
                Bash:
                    - set_inherit_options(True): works
                    - set_inherit_options(False): works
                Fish:
                    - set_inherit_options(True): works
                    - set_inherit_options(False): works
                Zsh:
                    - set_inherit_options(True): works
                    - set_inherit_options(False): works

        See Also:
            cli.CommandLine(..., inherit_options=BOOL, ...)
        '''

        _assert_is_bool(enable, "set_inherit_options", "enable")

        self.inherit_options = enable

    def set_option_stacking(self, enable):
        '''Sets wether short option stacking is allowed.

        Args:
            enable (bool):
                If True, option stacking is allowed.
                If False, it is not.

        Notes:
            This feature defaults to `True`.

            Implementation status for shells:
                Bash:
                    - set_option_stacking(True): works
                    - set_option_stacking(False): not implemented
                Fish:
                    - set_option_stacking(True): works
                    - set_option_stacking(False): works
                Zsh:
                    - set_option_stacking(True): works
                    - set_option_stacking(False): works

        '''

        _assert_is_bool(enable, "set_option_stacking", "enable")

        self.option_stacking = enable

    def set_long_option_argument_separator(self, separator):
        '''Sets which separators are used for delimiting a long option from
        its argument.

        Args:
            separator (string):
                If `space`, only the form `--option argument` is allowed.
                If `equals`, only the form `--option=argument` is allowed.
                If `both`, both forms are allowed.

        Notes:
            This feature defaults to `both`.

            Implementation status for shells:
                Bash:
                    - set_long_option_argument_separator('space'): works
                    - set_long_option_argument_separator('equals'): works
                    - set_long_option_argument_separator('both'): works
                Fish:
                    - set_long_option_argument_separator('space'): not implemented
                    - set_long_option_argument_separator('equals'): not implemented
                    - set_long_option_argument_separator('both'): works
                Zsh:
                    - set_long_option_argument_separator('space'): works
                    - set_long_option_argument_separator('equals'): works
                    - set_long_option_argument_separator('both'): works
        '''

        if separator not in ('space', 'equals', 'both'):
            raise AssertionError('Config.set_long_option_argument_separator: '
                                 'separator: Invalid value: '
                                 'expected: `space`, `equals` or `both`')

        self.long_opt_arg_sep = separator

    def set_vim_modeline(self, enable):
        '''Sets whether a vim modeline comment shall be appended to the code.

        The modeline comment looks like this:

            # vim: ft=zsh ts=2 sts=2 sw=2 et

        Args:
            enable (bool):
                If True, add a vim modline comment;
                if False, don't add a modline comment.

        Notes:
            This feature defaults to `True`.
        '''

        _assert_is_bool(enable, "set_vim_modeline", "enable")

        self.vim_modeline = enable

    def set_bash_completions_version(self, version):
        '''Sets the version of bash-completions.

        Args:
            version (tuple):
                A tuple in form like (2,) or (2, 12).

        Notes:
            This defaults to (2,).
        '''

        assert isinstance(version, tuple)

        self.bash_completions_version = version

    def set_zsh_compdef(self, enable):
        '''Sets whether a `#compdef` comment is written at the top of the
        generated zsh script.

        The `#compdef` directive is used by Zsh to automatically associate the
        generated completion file with a command, enabling autoload
        functionality.

        If you plan to load the Zsh completion file manually by sourcing it,
        omitting this line may be necessary.

        Args:
            enable (bool):
                If true, add a `#compdef` line on top of the file;
                If false, don't add a `#compdef` line.

        Notes:
            This feature defaults to `True`
        '''

        _assert_is_bool(enable, "set_zsh_compdef", "enable")

        self.zsh_compdef = enable

    def set_keep_comments(self, enable):
        '''Sets whether comments shall be included in the generated code.

        Args:
            enable (bool):
                If True, comments are included.
                if False, don't include comments.

        Notes:
            This feature defaults to `False`.
        '''

        _assert_is_bool(enable, "set_keep_comments", "enable")

        self.keep_comments = enable

    def set_line_length(self, length):
        '''Sets the maximum line length of the generated code.

        Args:
            length (int):
                The maximum line length. If zero or negative, maximum line
                length is selected.

        Notes:
            This defaults to `80`.
        '''

        assert isinstance(length, int)

        self.line_length = length if length > 0 else sys.maxsize

    def set_fish_fast(self, enable):
        '''Use faster conditions at the cost of correctness.'''

        _assert_is_bool(enable, "set_fish_fast", "enable")

        self.fish_fast = enable

    def set_fish_inline_conditions(self, enable):
        '''Don't store conditions in an extra variable.'''

        _assert_is_bool(enable, "set_fish_inline_conditions", "enable")

        self.fish_inline_conditions = enable

    def include_file(self, file):
        '''Add a file which should be included to the generated code.'''

        assert isinstance(file, str), \
            f"Config.include_file: file: expected str, got `{file}`"

        self.include_files.append(file)

    def include_many_files(self, files):
        '''Add files which should be included to the generated code.'''

        assert is_list_type(files)

        self.include_files.extend(files)

    def add_comments(self, comments):
        '''Add comments to the generated output.'''

        assert is_list_type(comments)

        self.comments.extend(comments)

    def disable_hidden(self, disable):
        '''Disable hidden options.

        This disables hidden options completely.
        '''

        _assert_is_bool(disable, "disable_hidden", "disable")

        self.disabled_hidden = disable

    def disable_final(self, disable):
        '''Disable final options.

        This disables final options completely.
        '''

        _assert_is_bool(disable, "disable_final", "disable")

        self.disabled_final = disable

    def disable_groups(self, disable):
        '''Disable option grouping.

        This disables the mutually exclusive feature for options completely.
        '''

        _assert_is_bool(disable, "disable_groups", "disable")

        self.disabled_groups = disable

    def disable_repeatable(self, disable):
        '''Disable repeatable options.

        This disables repeatable options completely.

        Despite its name, this function actually does the opposite:
        Instead of making all options non-repeatable, it makes all options
        repeatable.
        '''

        _assert_is_bool(disable, "disable_repeatable", "disable")

        self.disabled_repeatable = disable

    def disable_when(self, disable):
        '''Disable when feature.

        This disables conditional options and positionals completely.
        '''

        _assert_is_bool(disable, "disable_when", "disable")

        self.disabled_when = disable



#!/usr/bin/env python3

import os
import sys
import glob
import shutil
import subprocess
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('--dir', default='/tmp')
parser.add_argument('--package-dir', default='/tmp/crazy-complete-packages')
parser.add_argument('--ignore-errors', action='store_true')
parser.add_argument('--os', action='append',
    help='Only build for OS. Can be specified multiple times')
parser.add_argument('--dry', action='store_true',
    help='Only print out what would be built')

OPTS = parser.parse_args()

SCRIPT_PATH = os.path.abspath(sys.argv[0])
SCRIPT_DIR = os.path.dirname(SCRIPT_PATH)
os.chdir(SCRIPT_DIR)

if os.geteuid() != 0:
    raise Exception("Please run this script as root")

def run(*args):
    result = subprocess.run(
        args,
        stderr=sys.stderr,
        stdout=sys.stdout,
        check=False)

    if result.returncode != 0:
        raise Exception('Command `%s` failed' % ' '.join(args))

class ProjectBuild:
    def __init__(self, os, build_script, image, package_glob):
        self.os = os
        self.build_script = build_script
        self.image = image
        self.package_glob = package_glob
        self.directory = OPTS.dir
        self.chroot_dir = '%s/%s-crazy-complete-chroot' % (self.directory, self.os)
        self.podman_root = '%s/podman-root' % self.directory
        self.podman_runroot = '%s/podman-runroot' % self.directory

    def create_chroot(self):
        if os.path.exists(self.chroot_dir):
            print('Skip installing chroot %s ...' % self.chroot_dir)
            return

        print('Installing chroot %s ...' % self.chroot_dir)
        run('./create_chroot.sh', self.podman_root, self.podman_runroot, self.image, self.chroot_dir)

    def mount_chroot(self):
        shutil.copy('/etc/resolv.conf', '%s/etc/resolv.conf' % self.chroot_dir)

        run('mount', '-t', 'proc', 'proc', '%s/proc' % self.chroot_dir)

        run('mount', '--rbind', '/sys',    '%s/sys'  % self.chroot_dir)
        run('mount', '--make-rslave',      '%s/sys'  % self.chroot_dir)

        run('mount', '--rbind', '/dev',    '%s/dev'  % self.chroot_dir)
        run('mount', '--make-rslave',      '%s/dev'  % self.chroot_dir)

        run('mount', '--rbind', '/run',    '%s/run'  % self.chroot_dir)
        run('mount', '--make-rslave',      '%s/run'  % self.chroot_dir)

    def umount_chroot(self):
        try: run('umount', '-R', '%s/dev'  % self.chroot_dir)
        except: pass
        try: run('umount', '-R', '%s/sys'  % self.chroot_dir)
        except: pass
        try: run('umount', '-R', '%s/proc' % self.chroot_dir)
        except: pass
        try: run('umount', '-R', '%s/run'  % self.chroot_dir)
        except: pass

    def package(self):
        print('Building for %s ...' % self.os)
        script = './packaging/%s' % self.build_script
        dest = '%s/root' % self.chroot_dir
        print("Copying %s to %s" % (script, dest))
        shutil.copy(script, dest)
        run('chroot', self.chroot_dir, './root/%s' % self.build_script)

    def move_package(self):
        pattern = '%s/root/%s' % (self.chroot_dir, self.package_glob)
        files = glob.glob(pattern)
        if not files:
            raise Exception('No package found')
        if len(files) > 1:
            raise Exception('Too much packages found: %s' % files)
        package = files[0]
        package_basename = os.path.basename(package)
        print("Found package: ", package)
        dest_package = '%s/%s-%s' % (OPTS.package_dir, self.os, package_basename)
        print("Copying %s to %s" % (package, dest_package))
        shutil.copy(package, dest_package)

os.makedirs(OPTS.package_dir, exist_ok=True)

DEBIAN_TRIXIE     = "docker.io/library/debian:trixie"
DEBIAN_BOOKWORM   = "docker.io/library/debian:bookworm"

UBUNTU_NOBLE      = "ubuntu:noble"
UBUNTU_JAMMY      = "ubuntu:jammy"

LINUX_MINT_22     = "docker.io/linuxmintd/mint22-amd64"
LINUX_MINT_21     = "docker.io/linuxmintd/mint21-amd64"

FEDORA_44         = "registry.fedoraproject.org/fedora:44"
FEDORA_43         = "registry.fedoraproject.org/fedora:43"

OPENSUSE_IMAGE    = "registry.opensuse.org/opensuse/tumbleweed:latest"
ARCHLINUX_IMAGE   = "docker.io/library/archlinux:latest"

builds = [
    ProjectBuild("debian-trixie",   "debian.sh",     DEBIAN_TRIXIE,    "*.deb"),
    ProjectBuild("debian-bookworm", "debian.sh",     DEBIAN_BOOKWORM,  "*.deb"),
    ProjectBuild("ubuntu-noble",    "debian.sh",     UBUNTU_NOBLE,     "*.deb"),
    ProjectBuild("ubuntu-jammy",    "debian.sh",     UBUNTU_JAMMY,     "*.deb"),
    ProjectBuild("linux-mint-22",   "debian.sh",     LINUX_MINT_22,    "*.deb"),
    ProjectBuild("linux-mint-21",   "debian.sh",     LINUX_MINT_21,    "*.deb"),
    ProjectBuild("fedora-44",       "fedora.sh",     FEDORA_44,        "*.rpm"),
    ProjectBuild("fedora-43",       "fedora.sh",     FEDORA_43,        "*.rpm"),
    ProjectBuild("opensuse",        "opensuse.sh",   OPENSUSE_IMAGE,   "*.rpm"),
    ProjectBuild("arch-linux",      "arch-linux.sh", ARCHLINUX_IMAGE,  "*.pkg.tar.zst"),
]

selected_builds = builds

if OPTS.os:
    selected_builds = filter(lambda rule: rule.os in OPTS.os, selected_builds)

for build in selected_builds:
    if OPTS.dry:
        print('Building for:', build.os)
        continue

    try:
        build.create_chroot()
    except Exception as e:
        if OPTS.ignore_errors:
            continue
        else:
            raise e

    try:
        build.mount_chroot()
        build.package()
        build.move_package()
    except Exception as e:
        if OPTS.ignore_errors:
            continue
        else:
            raise e
    finally:
        build.umount_chroot()



import time

def strip_lines(string):
    return '\n'.join(s.rstrip() for s in string.split('\n')).rstrip()

class TerminalBase:
    def get_output_stripped(self):
        return strip_lines(self.get_output())

    def clear_screen(self):
        result = ''
        while result != '>':
            # Cancel potentially running task
            self.send_ctrl("c")
            # Clear current command line, as fish does not do it on ^C
            self.send_ctrl("u")
            self.wait_for_last_line('>', 5, 0.01)
            self.send('clear')
            self.wait_for_last_line('> clear', 5, 0.01)
            self.send_line('')
            result = self.wait_for_text('>', 5, 0.01)

    def complete(self, commandline, num_tabs=1, wait=5, expected=None, fast=False):
        self.clear_screen()
        self.send(commandline)

        for _ in range(num_tabs):
            self.send_tab()

        if fast:
            return self.wait_for_text(expected, wait, 0.01)
        else:
            time.sleep(wait)
            return self.get_output_stripped()

    def wait_for_text(self, expected, timeout, poll_interval):
        expected = strip_lines(expected)

        while timeout > 0.0:
            result = self.get_output_stripped()
            if result == expected:
                return result
            timeout -= poll_interval
            time.sleep(poll_interval)
        return self.get_output_stripped()

    def wait_for_last_line(self, expected, timeout, poll_interval):
        expected = expected.rstrip()

        while timeout > 0.0:
            result = self.get_output_stripped().split('\n')[-1]
            if result == expected:
                return True
            timeout -= poll_interval
            time.sleep(poll_interval)
        return False



#!/usr/bin/env python3

from setuptools import setup

setup(
    name='crazy-complete',
    version='0.3.8',
    author='Benjamin Abendroth',
    author_email='braph93@gmx.de',
    packages=['crazy_complete'],
    scripts=['crazy-complete'],
    description='Generate shell completion files for all major shells',
    url='https://github.com/crazy-complete/crazy-complete',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Unix Shell',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: User Interfaces',
        'Topic :: Software Development :: Code Generators',
        'Topic :: System :: Shells',
        'Topic :: Utilities',
    ],
    license='GPL-3.0',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    python_requires='>=3.0',
)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Type utility functions.'''

from types import NoneType

from .errors import CrazyTypeError


def is_dict_type(obj):
    '''Return if `obj` is a dictionary type.'''

    return hasattr(obj, 'items')


def is_list_type(obj):
    '''Return if `obj` is list or tuple.'''

    return isinstance(obj, (list, tuple))


def validate_type(value, types, parameter_name):
    '''Check if `value` is one of `types`.'''

    if not isinstance(value, types):
        types_strings = []
        for t in types:
            types_strings.append({
                str:      'str',
                int:      'int',
                float:    'float',
                list:     'list',
                tuple:    'tuple',
                dict:     'dict',
                bool:     'bool',
                NoneType: 'None'}[t])
        types_string = '|'.join(types_strings)

        raise CrazyTypeError(parameter_name, types_string, value)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Bash utility functions.'''

from . import shell


def make_option_variable_name(option, prefix=''):
    '''Make a variable for an option.'''
    long_options = option.get_long_option_strings()
    if long_options:
        return prefix + shell.make_identifier(long_options[0].lstrip('-'))

    old_options = option.get_old_option_strings()
    if old_options:
        return prefix + shell.make_identifier(old_options[0].lstrip('-'))

    short_options = option.get_short_option_strings()
    if short_options:
        return prefix + short_options[0][1]

    raise AssertionError("make_option_variable_name: Should not be reached")


class VariableManager:
    '''Variable manager.'''

    def __init__(self, prefix):
        self.prefix = prefix
        self.variables = set()

    def make_variable(self, option):
        '''Make a variable for an option.'''

        var = make_option_variable_name(option, self.prefix)
        self.variables.add(var)
        return var

    def capture_variable(self, option):
        '''Return the capture variable of an option, creating it if necessary.'''

        if option.capture is None:
            option.capture = self.make_variable(option)
        return option.capture

    def get_variables(self):
        '''Get a list of all defined variables.'''

        return list(sorted(self.variables))


class CasePatterns:
    '''Functions for creating case patterns.'''

    @staticmethod
    def for_long_without_arg(option_strings):
        '''Return a case pattern for long options without arguments.'''

        return '|'.join(option_strings)

    @staticmethod
    def for_long_with_arg(option_strings):
        '''Return a case pattern for long options with arguments.'''

        return '|'.join(f'{o}=*' for o in option_strings)

    @staticmethod
    def for_short(option_strings):
        '''Return a case pattern for short options.'''

        return '|'.join(o[1] for o in option_strings)

    @staticmethod
    def for_old_without_arg(option_strings):
        '''Return a case pattern for old options without arguments.'''

        if len(option_strings) <= 2:
            return '|'.join(option_strings)

        return '-@(%s)' % '|'.join(o.lstrip('-') for o in option_strings)


def make_file_extension_pattern(extensions, fuzzy):
    '''Make a case-insensitive glob pattern matching `extensions`.

    Takes a list of extensions (e.g. ['txt', 'jpg']) and returns a Bash-style
    pattern that matches all of them, ignoring case.

    Example output:
        '@([tT][xX][tT]|[jJ][pP][gG])'
    '''

    patterns = []

    for extension in extensions:
        pattern = ''
        for c in extension:
            if c.isalpha():
                pattern += '[%s%s]' % (c.lower(), c.upper())
            else:
                pattern += c

        patterns.append(pattern)

    if not fuzzy:
        return '@(%s)' % '|'.join(patterns)

    return '@(%s)*' % '|'.join(patterns)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains code for completing option strings in Bash.'''

from collections import namedtuple

from . import cli
from . import algo
from . import shell
from . import bash_when
from .str_utils import indent


_Option = namedtuple('_Option', ('option', 'conditions', 'when'))


def _make_option_strings(options):
    opt_strings = []

    for opt in options:
        opt_strings.extend(opt.get_short_option_strings())

        add_equal_sign = (
            opt.has_optional_arg() or (
                opt.has_required_arg() and
                opt.long_opt_arg_sep in ('both', 'equals')
            )
        )

        if add_equal_sign:
            opt_strings.extend(f'{s}=' for s in opt.get_long_option_strings())
            opt_strings.extend(f'{s}=' for s in opt.get_old_option_strings())
        else:
            opt_strings.extend(opt.get_long_option_strings())
            opt_strings.extend(opt.get_old_option_strings())

    return shell.join_quoted(opt_strings)


def _get_option_full_condition(option):
    if option.conditions and option.when:
        return '(( %s )) && %s' % (' && '.join(option.conditions), option.when)
    if option.conditions:
        return '(( %s ))' % (' && '.join(option.conditions))
    if option.when:
        return '%s' % option.when

    return ''


def _generate_option_strings_completion(options):
    r = []

    grouped_by_condition = algo.group_by(options, _get_option_full_condition)
    for condition, opts in grouped_by_condition.items():
        s = 'opts+=(%s)' % _make_option_strings([o.option for o in opts])
        if condition:
            s = '%s && %s' % (condition, s)
        r.append(s)

    return '\n'.join(r)


def _generate_final_check_with_options(final_conditions, options):
    option_strings_completion = _generate_option_strings_completion(options)

    if not final_conditions:
        return option_strings_completion

    r  = 'if (( %s )); then\n' % ' && '.join(final_conditions)
    r += '%s\n' % indent(option_strings_completion, 2)
    r += 'fi'
    return r


def generate(generator):
    '''Generate option strings completion code.'''

    if len(generator.options) == 0:
        return None

    commandline = generator.commandline
    variable_manager = generator.variable_manager
    options = []
    final_conditions = []

    for final_option in commandline.get_final_options():
        final_conditions += ["! ${#%s[@]}" % variable_manager.capture_variable(final_option)]

    for option in commandline.options:
        if option.hidden:
            continue

        conditions = []

        for exclusive_option in option.get_conflicting_options():
            conditions += ["! ${#%s[@]}" % variable_manager.capture_variable(exclusive_option)]

        if not option.repeatable:
            conditions += ["! ${#%s[@]}" % variable_manager.capture_variable(option)]

        for final_condition in final_conditions:
            try:
                conditions.remove(final_condition)
            except ValueError:
                pass

        conditions = algo.uniq(sorted(conditions))

        when = None
        if option.when is not None:
            when = bash_when.generate_when_conditions(commandline, variable_manager, option.when)

        options.append(_Option(option, conditions, when))

    r  = 'if (( ! END_OF_OPTIONS )) && [[ "$cur" = -* ]]; then\n'
    r += '  local -a opts\n'
    r += '%s\n' % indent(_generate_final_check_with_options(final_conditions, options), 2)
    r += '  COMPREPLY+=($(compgen -W "${opts[*]}" -- "$cur"))\n'
    r += '  [[ ${COMPREPLY-} == *= ]] && compopt -o nospace\n'
    r += '  return 1\n'
    r += 'fi'

    return r
